/* -*- Mode: C++; -*- */
/* VER: $Id: ANN.c,v 1.11 2003/09/05 01:43:35 olethros Exp $ */

#include "ANN.h"

//==========================================================
// NewANN
//----------------------------------------------------------
#undef ANN_DBUG

ANN *NewANN(int n_inputs, int n_outputs)
{
	ANN *ann = NULL;

	if (!(ann = AllocM(ANN, 1))) {
		Serror("Could not allocate ANN\n");
		return NULL;
	}
	ann->x = NULL;
	ann->y = NULL;
	ann->t = NULL;
	ann->d = NULL;
	ann->error = NULL;
	ann->c = NULL;
	ann->a = 0.1f;
	ann->n_inputs = n_inputs;
	ann->n_outputs = n_outputs;
	ann->batch_mode = false;

	/* outputs are not allocated */

	if (!(ann->error = AllocM(real, n_outputs))) {
		Serror("Could not allocate errors\n");
		DeleteANN(ann);
		return NULL;
	}

	if (!(ann->d = AllocM(real, n_outputs))) {
		Serror("Could not allocate derivatives\n");
		DeleteANN(ann);
		return NULL;
	}

	if (!(ann->c = List())) {
		Serror("Could not allocate list\n");
		DeleteANN(ann);
		return NULL;
	}
#ifdef ANN_DBUG
	message("Creating ANN with %d inputs and %d outputs", n_inputs,
		n_outputs);
#endif
	return ann;
}

//==========================================================
// DeleteANN
//----------------------------------------------------------
// Delete a neural network
int DeleteANN(ANN * ann)
{
	if (!ann) {
		Swarning("Attempting to delete NULL ANN\n");
		return DEC_ARG_INVALID;
	}

	if (ann->error) {
		FreeM(ann->error);
	}
	//  if (ann->x) {
	//    FreeM (ann->x);
	//  }

	if (ann->d) {
		FreeM(ann->d);
	}

	/* We must clear all allocations in the list */
	if (ann->c) {
		ClearList(ann->c);
		ann->c = NULL;
	}

	FreeM(ann);
	return 0;
}


//==========================================================
// ANN_AddHiddenLayer()
//----------------------------------------------------------
int ANN_AddHiddenLayer(ANN * ann, int n_nodes)
{
#ifdef ANN_DBUG
	message("Adding Hidden layer with %d nodes", n_nodes);
#endif

	LISTITEM *item = LastListItem(ann->c);
	if (item) {
		Layer *p = (Layer *) item->obj;
		ANN_AddLayer(ann, p->n_outputs, n_nodes, p->y);
	} else {
		ANN_AddLayer(ann, ann->n_inputs, n_nodes, ann->x);
	}
	return 0;
}

//==========================================================
// ANN_AddHiddenLayer()
//----------------------------------------------------------
int ANN_AddRBFHiddenLayer(ANN * ann, int n_nodes)
{
#ifdef ANN_DBUG
	message("Adding Hidden layer with %d nodes", n_nodes);
#endif
	LISTITEM *item = LastListItem(ann->c);
	if (item) {
		Layer *p = (Layer *) item->obj;
		ANN_AddRBFLayer(ann, p->n_outputs, n_nodes, p->y);
	} else {
		ANN_AddRBFLayer(ann, ann->n_inputs, n_nodes, ann->x);
	}
	return 0;
}



//==========================================================
// ANN_AddLayer()
//----------------------------------------------------------
Layer *ANN_AddLayer(ANN * ann, int n_inputs, int n_outputs, real * x)
{
	Layer *l = NULL;
	if ((x == NULL) && (ann->c->n)) {
		Swarning
		    ("Layer connects to null but layer list is not empty\n");
	}


	if (!(l = AllocM(Layer, 1))) {
		Serror("Could not allocate layer structure\n");
		return NULL;
	}

	assert(n_inputs > 0);
	assert(n_outputs > 0);

	l->n_inputs = n_inputs;
	l->n_outputs = n_outputs;
	l->x = x;
	l->a = ann->a;
	l->forward = &ANN_CalculateLayerOutputs;
	l->backward = &ANN_Backpropagate;
	l->f = &htan;
	l->f_d = &htan_d;

	if (!(l->y = AllocM(real, n_outputs))) {
		Serror("Could not allocate layer outputs\n");
		ANN_FreeLayer(l);
		return NULL;
	}

	if (!(l->z = AllocM(real, n_outputs))) {
		Serror("Could not allocate layer activations\n");
		ANN_FreeLayer(l);
		return NULL;
	}


	if (!(l->d = AllocM(real, n_inputs + 1 /*bias */ ))) {
		Serror("Could not allocate layer outputs\n");
		ANN_FreeLayer(l);
		return NULL;
	}

	if (!
	    (l->c =
	     AllocM(Connection, (n_inputs + 1 /*bias */ ) * n_outputs))) {
		Serror("Could not allocate connections\n");
		ANN_FreeLayer(l);
		return NULL;
	}

	l->rbf = NULL;

	real bound = 2.0f / sqrt((real) n_inputs);
	for (int i = 0; i < n_inputs + 1 /*bias */ ; i++) {
		Connection *c = &l->c[i * n_outputs];
		for (int j = 0; j < n_outputs; j++) {
			c->w = (drand48() - 0.5f) * bound;;
			c->c = 1;
			c->e = 0.0f;
			c->dw = 0.0f;
			c++;
		}
	}
	ListAppend(ann->c, (void *) l, &ANN_FreeLayer);
	return l;
}


//==========================================================
// ANN_AddRBFLayer()
//----------------------------------------------------------
Layer *ANN_AddRBFLayer(ANN * ann, int n_inputs, int n_outputs, real * x)
{
	Layer *l = NULL;
	if ((x == NULL) && (ann->c->n)) {
		Swarning
		    ("Layer connects to null and layer list not empty\n");
	}

	if (!(l = AllocM(Layer, 1))) {
		Serror("Could not allocate layer structure\n");
		return NULL;
	}

	assert(n_inputs > 0);
	assert(n_outputs > 0);

	l->n_inputs = n_inputs;
	l->n_outputs = n_outputs;
	l->x = x;
	l->a = ann->a;
	l->forward = &ANN_RBFCalculateLayerOutputs;
	l->backward = &ANN_RBFBackpropagate;
	l->f = &Exp;
	l->f_d = &Exp_d;

	if (!(l->y = AllocM(real, n_outputs))) {
		Serror("Could not allocate layer outputs\n");
		ANN_FreeLayer(l);
		return NULL;
	}

	if (!(l->z = AllocM(real, n_outputs))) {
		Serror("Could not allocate layer activations\n");
		ANN_FreeLayer(l);
		return NULL;
	}

	if (!(l->d = AllocM(real, n_inputs + 1 /*bias */ ))) {
		Serror("Could not allocate layer outputs\n");
		ANN_FreeLayer(l);
		return NULL;
	}

	if (!
	    (l->rbf =
	     AllocM(RBFConnection,
		    (n_inputs + 1 /*bias */ ) * n_outputs))) {
		Serror("Could not allocate connections\n");
		ANN_FreeLayer(l);
		return NULL;
	}

	l->c = NULL;

	real bound = 2.0f / sqrt((real) n_inputs);
	for (int i = 0; i < n_inputs + 1 /*bias */ ; i++) {
		RBFConnection *c = &l->rbf[i * n_outputs];
		for (int j = 0; j < n_outputs; j++) {
			c->w = (drand48() - 0.5f) * bound;;
			c->m = (drand48() - 0.5f) * 2.0f;
			c++;
		}
	}
	ListAppend(ann->c, (void *) l, &ANN_FreeLayer);
	return l;
}


//==========================================================
// ANN_FreeLayer()
//----------------------------------------------------------
void ANN_FreeLayer(void *l)
{
	ANN_FreeLayer((Layer *) l);
}

//==========================================================
// ANN_FreeLayer()
//----------------------------------------------------------
void ANN_FreeLayer(Layer * l)
{
	FreeM(l->y);
	if (l->c) {
		FreeM(l->c);
	}
	if (l->rbf) {
		FreeM(l->rbf);
	}
	FreeM(l->d);
	FreeM(l);

}

//==========================================================
// ANN_Init()
//----------------------------------------------------------
int ANN_Init(ANN * ann)
{
	// Add output layer
	LISTITEM *item = LastListItem(ann->c);
	Layer *l = NULL;
#ifdef ANN_DBUG
	message("Initialising");
#endif
	if (item) {
		Layer *p = (Layer *) item->obj;
		l = ANN_AddLayer(ann, p->n_outputs, ann->n_outputs, p->y);
	} else {
		l = ANN_AddLayer(ann, ann->n_inputs, ann->n_outputs,
				 ann->x);
	}
	if (l == NULL) {
		Serror("Could not create final layer\n");
		DeleteANN(ann);
		return -1;
	}
	ann->y = l->y;
	l->f = &linear;
	l->f_d = &linear_d;
	//  ann->t = l->t;
	return 0;
}



//==========================================================
// ANN_Reset()
//----------------------------------------------------------
void ANN_Reset(ANN * ann)
{
}


//==========================================================
// ANN_Input()
//----------------------------------------------------------
real ANN_Input(ANN * ann, real * x)
{
	LISTITEM *p = FirstListItem(ann->c);
	Layer *first_layer = (Layer *) p->obj;
	ann->x = x;
	first_layer->x = x;	// Setup input of first layer
	//  printf ("II: %f\n", ann->x[0]);
	while (p) {
		Layer *current_layer = (Layer *) p->obj;
		//    printf ("\tIII: %f\n", current_layer->x[0]);
		current_layer->forward(current_layer);
		p = NextListItem(ann->c);
	}
	return 0.0f;
}

//==========================================================
// ANN_CalculateLayerOutputs()
//----------------------------------------------------------
void ANN_CalculateLayerOutputs(Layer * current_layer)
{
	int i, j;
	int n_inputs = current_layer->n_inputs;
	int n_outputs = current_layer->n_outputs;
	real *x = current_layer->x;
	real *y = current_layer->y;
	real *z = current_layer->z;
	Connection *c;


	for (j = 0; j < n_outputs; j++) {
		z[j] = 0.0f;
	}

	c = current_layer->c;
	for (i = 0; i < n_inputs; i++) {
		for (j = 0; j < n_outputs; j++) {
			z[j] += x[i] * c->w;
			c++;
		}
	}

	// bias
	for (j = 0; j < n_outputs; j++) {
		z[j] += c->w;
		c++;
	}

	for (j = 0; j < n_outputs; j++) {
		y[j] = current_layer->f(z[j]);
	}
}

//==========================================================
// ANN_RBFCalculateLayerOutputs()
//----------------------------------------------------------
void ANN_RBFCalculateLayerOutputs(Layer * current_layer)
{
	int i, j;
	int n_inputs = current_layer->n_inputs;
	int n_outputs = current_layer->n_outputs;
	real *x = current_layer->x;
	real *y = current_layer->y;
	real *z = current_layer->z;
	RBFConnection *c;


	for (j = 0; j < n_outputs; j++) {
		z[j] = 0.0f;
	}

	c = current_layer->rbf;
	for (i = 0; i < n_inputs; i++) {
		real in = x[i];
		for (j = 0; j < n_outputs; j++, c++) {
			real o = (in - c->m) * c->w;
			z[j] += o * o;
		}
	}

	for (j = 0; j < n_outputs; j++) {
		z[j] = -0.5f * z[j];
		y[j] = current_layer->f(z[j]);
	}
}


//==========================================================
// ANN_Train()
//----------------------------------------------------------
real ANN_Train(ANN * ann, real * x, real * t)
{
	LISTITEM *p = LastListItem(ann->c);
	Layer *l = (Layer *) p->obj;
	real sum = 0.0f;
	int j;

	ANN_Input(ann, x);

	for (j = 0; j < ann->n_outputs; j++) {
		real f = l->f_d(ann->y[j]);
		real e = t[j] - ann->y[j];
		ann->d[j] = e * f;
		sum += e * e;
	}

	l->backward(p, ann->d);

	return sum;
}



//==========================================================
// ANN_Backpropagate
//----------------------------------------------------------
/* d are the derivatives at the outputs. */
real ANN_Backpropagate(LISTITEM * p, real * d)
{
	int i, j;
	float f;
	float a;
	Layer *l = (Layer *) p->obj;
	LISTITEM *back = p->prev;
	Layer *back_layer = NULL;
	a = l->a;
	//  message ("backing with in: %d",l->x);
	if (back) {
		//message ("and to prev");
		back_layer = (Layer *) back->obj;
		for (i = 0; i < l->n_inputs; i++) {
			l->d[i] = 0.0f;
			Connection *c = &l->c[i * l->n_outputs];
			for (j = 0; j < l->n_outputs; j++) {
				l->d[i] += c->w * d[j];
				c++;
			}
			f = back_layer->f_d(l->x[i]);
			l->d[i] = l->d[i] * f;
		}

		/* bias */
		i = l->n_inputs;
		l->d[i] = 0.0f;
		Connection *c = &l->c[i * l->n_outputs];
		for (j = 0; j < l->n_outputs; j++) {
			l->d[i] += c->w * d[j];
			c++;
		}
		f = back_layer->f_d(1.0f);
		l->d[i] = l->d[i] * f;

		back_layer->backward(back, l->d);
	}
	//update weights
	for (i = 0; i < l->n_inputs; i++) {
		Connection *c = &l->c[i * l->n_outputs];
		real dx = a * l->x[i];
		if (l->batch_mode) {
			for (j = 0; j < l->n_outputs; j++) {
				c->dw = dx * d[j];
				c++;
			}
		} else {
			for (j = 0; j < l->n_outputs; j++) {
				c->w += dx * d[j];
				c++;
			}
		}
	}
	// update bias weight
	{
		Connection *c = &l->c[l->n_inputs * l->n_outputs];
		if (l->batch_mode) {
			for (j = 0; j < l->n_outputs; j++) {
				c->dw = a * d[j];
				c++;
			}
		} else {
			for (j = 0; j < l->n_outputs; j++) {
				c->dw = a * d[j];
				c++;
			}
		}
	}
	return 0.0f;
}

//==========================================================
// ANN_RBFBackpropagate
//----------------------------------------------------------
real ANN_RBFBackpropagate(LISTITEM * p, real * d)
{
	int i, j;
	float f;
	float a;
	Layer *l = (Layer *) p->obj;
	LISTITEM *back = p->prev;
	Layer *back_layer = NULL;
	a = l->a;

	if (back) {
		back_layer = (Layer *) back->obj;
		for (i = 0; i < l->n_inputs; i++) {
			l->d[i] = 0.0f;
			RBFConnection *c = &l->rbf[i * l->n_outputs];
			for (j = 0; j < l->n_outputs; j++) {
				real dx = l->x[i] - c->m;
				real dm = d[j] * dx * c->w * c->w;
				l->d[j] -= dm;
				c++;
			}
			f = back_layer->f_d(l->x[i]);
			l->d[i] = l->d[i] * f;
		}
		back_layer->backward(back, l->d);
	}

	return 0.0f;
	//update weights
	for (i = 0; i < l->n_inputs; i++) {
		RBFConnection *c = &l->rbf[i * l->n_outputs];
		real dx = l->x[i] - c->m;
		for (j = 0; j < l->n_outputs; j++) {
			real dy = d[j];
			real dx2 = a * dy * dx * c->w;
			real dm = dx2 * c->w;
			real dw = dx2 * dx;
			c->m += dm;
			c->w += dw;
			c++;
		}
	}
	return 0.0f;
}


void ANN_LayerBatchAdapt(Layer * l)
{
	int i, j;

	if (l->batch_mode == false) {
		Serror("Batch adapt yet not in batch mode!");
	}
	//update weights
	for (i = 0; i < l->n_inputs; i++) {
		Connection *c = &l->c[i * l->n_outputs];
		for (j = 0; j < l->n_outputs; j++) {
			c->w += c->dw;
			c++;
		}
	}
	// update bias weight
	{
		Connection *c = &l->c[l->n_inputs * l->n_outputs];
		for (j = 0; j < l->n_outputs; j++) {
			c->w += c->dw;
			c++;
		}
	}
}

//==========================================================
// ANN_Test()
//----------------------------------------------------------
real ANN_Test(ANN * ann, real * x, real * t)
{
	LISTITEM *p = LastListItem(ann->c);
	Layer *l = (Layer *) p->obj;
	real sum = 0.0f;
	int j;
	ANN_Input(ann, x);

	for (j = 0; j < ann->n_outputs; j++) {
		real f = l->f_d(ann->y[j]);
		real e = t[j] - ann->y[j];
		ann->d[j] = e * f;
		sum += e * e;
	}
	return sum;
}

//==========================================================
// ANN_GetOutput()
//----------------------------------------------------------
real *ANN_GetOutput(ANN * ann)
{
	return ann->y;
}

//==========================================================
// ANN_GetError()
//----------------------------------------------------------
real ANN_GetError(ANN * ann)
{
	return 0.0f;
}

//==========================================================
// ANN_SetLearningRate()
//----------------------------------------------------------
void ANN_SetLearningRate(ANN * ann, real a)
{
	LISTITEM *c;

	ann->a = a;
	c = FirstListItem(ann->c);
	while (c) {
		Layer *l = (Layer *) c->obj;
		l->a = a;
		c = NextListItem(ann->c);
	}
}

//==========================================================
// ANN_SetBatchMode
//----------------------------------------------------------
void ANN_SetBatchMode(ANN * ann, bool batch)
{
	LISTITEM *c;

	ann->batch_mode = batch;
	c = FirstListItem(ann->c);
	while (c) {
		Layer *l = (Layer *) c->obj;
		l->batch_mode = batch;
		c = NextListItem(ann->c);
	}
}

//==========================================================
// ANN_BatchAdapt
//----------------------------------------------------------
void ANN_BatchAdapt(ANN * ann)
{
	LISTITEM *c;

	c = FirstListItem(ann->c);
	while (c) {
		Layer *l = (Layer *) c->obj;
		ANN_LayerBatchAdapt(l);
		c = NextListItem(ann->c);
	}
}



//==========================================================
// ANN_ShowWeights()
//----------------------------------------------------------
real ANN_ShowWeights(ANN * ann)
{
	LISTITEM *c;
	real sum = 0.0f;
	c = FirstListItem(ann->c);
	while (c) {
		Layer *l = (Layer *) c->obj;
		sum += ANN_LayerShowWeights(l);
		c = NextListItem(ann->c);
	}
	return sum;
}

//==========================================================
// ANN_LayerShowWeights()
//----------------------------------------------------------
real ANN_LayerShowWeights(Layer * l)
{
	int i, j;
	Connection *c = l->c;
	real sum = 0.0f;
	for (i = 0; i < l->n_inputs + 1 /*bias */ ; i++) {
		for (j = 0; j < l->n_outputs; j++) {
			sum += (c->w) * (c->w);	//printf ("%f ", c->w);
			printf("#%f ", c->w);
			c++;
		}
	}
	printf("\n");
	//  printf ("%f ", sum);
	return sum;
}


//==========================================================
// ANN_ShowInputs()
//----------------------------------------------------------
real ANN_ShowInputs(ANN * ann)
{
	LISTITEM *c;
	real sum = 0.0f;
	c = FirstListItem(ann->c);
	while (c) {
		Layer *l = (Layer *) c->obj;
		sum += ANN_LayerShowInputs(l);
		c = NextListItem(ann->c);
	}
	return sum;
}

//==========================================================
// ANN_LayerShowInputs()
//----------------------------------------------------------
real ANN_LayerShowInputs(Layer * l)
{
	int i;
	real sum = 0.0f;
	for (i = 0; i < l->n_inputs; i++) {
		printf("#%f ", l->x[i]);
	}
	printf("-->");
	for (i = 0; i < l->n_outputs; i++) {
		printf("#(%f)%f ", l->f(l->z[i]), l->y[i]);
	}

	printf("\n");
	return sum;
}


//==========================================================
// ANN_ShowOutputs()
//----------------------------------------------------------
void ANN_ShowOutputs(ANN * ann)
{
	int i;

	for (i = 0; i < ann->n_outputs; i++) {
		printf("%f ", ann->y[i]);
	}
	printf("\n");
}



//==========================================================
// ANN_SetOutputsToLinear()
//----------------------------------------------------------
void ANN_SetOutputsToLinear(ANN * ann)
{
	LISTITEM *c;
	c = LastListItem(ann->c);
	if (c) {
		Layer *l = (Layer *) c->obj;
		l->f = &linear;
		l->f_d = &linear_d;
	} else {
		Serror("Could not set outputs to linear\n");
	}
}

//==========================================================
// ANN_SetOutputsToTanH()
//----------------------------------------------------------
void ANN_SetOutputsToTanH(ANN * ann)
{
	LISTITEM *c;
	c = LastListItem(ann->c);
	if (c) {
		Layer *l = (Layer *) c->obj;
		l->f = &htan;
		l->f_d = &htan_d;
	} else {
		Serror("Could not set outputs to TanH\n");
	}
}


//==========================================================
// Exp()
//----------------------------------------------------------
real Exp(real x)
{
	return exp(x);
}

//==========================================================
// Exp_d()
//----------------------------------------------------------
real Exp_d(real x)
{
	return x;
}

//==========================================================
// htan()
//----------------------------------------------------------
real htan(real x)
{
	return tanh(x);
}

//==========================================================
// htan_d()
//----------------------------------------------------------
real htan_d(real x)
{
	real f = tanh(x);
	return (1.0 - f * f);
}

//==========================================================
// linear()
//----------------------------------------------------------
real linear(real x)
{
	return x;
}

//==========================================================
// linear_d()
//----------------------------------------------------------
real linear_d(real x)
{
	return 1.0f;
}
