/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* VER: $Id: ANN.h,v 1.7 2003/08/11 10:07:22 olethros Exp $ */

#ifndef ANN_H
#define ANN_H

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "debug.h"
#include "util.h"
#include "string_utils.h"
#include "List.h"
#include "real.h"


typedef struct Connection_ {
  int c; // connected?
  real w; // weight
  real dw; // Weight-change
  real e; // eligibility;
} Connection;

typedef struct RBFConnection_ {
  real w; // weight (=1/sigma)
  real m; // mean
} RBFConnection;


typedef struct Layer_ {
  int n_inputs;
  int n_outputs;
  real* x; //inputs;
  real* y; //outputs
  real* z; //activation
  real* d; //derivatives
  Connection* c; //connections
  RBFConnection* rbf;
  real a; //learning rate
  bool batch_mode; //do not update weights immediately
  void (*forward) (struct Layer_* current_layer);
  real (*backward) (LISTITEM* p, real* d);
  real (*f) (real x); //activation function
  real (*f_d) (real x); //derivative of activation function
} Layer;

typedef struct ANN_ {
  int n_inputs;
  int n_outputs;
  LIST* c; //connection layers
  real* x; // unit inputs
  real* y; // unit activations
  real* t; // targets
  real* d;
  real a; // learning rate
  real* error; // errors
  bool batch_mode;
} ANN;


/**************** User-level API ******************/
/* Object Management Interface */
ANN* NewANN(int n_inputs, int n_outputs);
int DeleteANN(ANN* ann);

/* Setup Inteface */
int ANN_AddHiddenLayer(ANN* ann, int n_nodes);
int ANN_AddRBFHiddenLayer (ANN* ann, int n_nodes);
int ANN_Init(ANN* ann);
void ANN_SetOutputsToTanH(ANN* ann);
void ANN_SetOutputsToLinear(ANN* ann);
void ANN_SetLearningRate(ANN* ann, real a);
void ANN_Reset(ANN* ann);

/* Functionality Interface */
real ANN_Input(ANN* ann, real* x);
real ANN_Train(ANN* ann, real* x, real* t);
void ANN_SetBatchMode(ANN* ann, bool batch);
void ANN_BatchAdapt(ANN* ann);
real ANN_Test(ANN* ann, real* x, real* t);
real* ANN_GetOutput(ANN* ann);
real ANN_GetError(ANN* ann);

/********* Low-level code **********/

/* Sub-object management functions */
Layer* ANN_AddLayer (ANN* ann, int n_inputs, int n_outputs, real* x);
Layer* ANN_AddRBFLayer (ANN* ann, int n_inputs, int n_outputs, real* x);
void ANN_FreeLayer (void* l);
void ANN_FreeLayer (Layer* l);

/* Calculations */
void ANN_CalculateLayerOutputs (Layer* current_layer);
real ANN_Backpropagate (LISTITEM* p, real* d);
void ANN_RBFCalculateLayerOutputs (Layer* current_layer);
real ANN_RBFBackpropagate (LISTITEM* p, real* d);
void ANN_LayerBatchAdapt (Layer* l);

/* Output functions and derivatives */
real Exp (real x);
real Exp_d (real x);
real htan (real x);
real htan_d (real x);
real linear (real x);
real linear_d (real x);

/* Debugging functions */
real ANN_LayerShowWeights (Layer* l);
real ANN_ShowWeights(ANN* ann);
void ANN_ShowOutputs(ANN* ann);
real ANN_ShowInputs(ANN* ann);
real ANN_LayerShowInputs(Layer* l);
#endif /* ANN_H */
