/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* VER: $Id: ISOLearning.c,v 1.1 2003/07/29 01:21:51 olethros Exp $ */

#include "ISOLearning.h"

//==========================================================
// 
//----------------------------------------------------------
ISOLearner* NewISOLearner (int n_inputs, int n_outputs, int n_filters, real dt)
{
	ISOLearner* iso = NULL;
	if (!(iso = AllocM (ISOLearner, 1))) {
		return NULL;
	}
	
	iso->x = NULL;
	iso->y = NULL;
	iso->h = NULL;
	iso->w = NULL;
	iso->dt = dt;

	if ((n_inputs <=0)||(n_outputs <=0)||(n_filters <=0)) {
		FreeISOLearner (iso);
	}

	iso->n_inputs = n_inputs;
	iso->n_outputs = n_outputs;
	iso->n_filters = n_filters;
	if ((iso->x = AllocM (real, n_inputs))==NULL) {
		FreeISOLearner (iso);
		return NULL;
	}
	if ((iso->y = AllocM (real, n_outputs))==NULL) {
		FreeISOLearner (iso);
		return NULL;
	}
	if ((iso->w = AllocM (real, n_filters*n_outputs))==NULL) {
		FreeISOLearner (iso);
		return NULL;
	}
	if ((iso->h = AllocM (ISOFilter, n_filters))==NULL) {
		FreeISOLearner (iso);
		return NULL;
	}
	for (int i=0; i<n_filters; i++) {
		if (NewISOFilter (&iso->h[i], n_inputs)) {
			Serror ("Could not create filter");
			FreeISOLearner (iso);
			return NULL;
		}
	}
	//	printf ("Created ISO Learner\n");
	return iso;
}


//==========================================================
// 
//----------------------------------------------------------
int FreeISOLearner (ISOLearner* iso)
{
	//	printf ("Freeing ISO\n");
	if (iso) {
		if (iso->x) {
			FreeM (iso->x);
		}
		if (iso->y) {
			FreeM (iso->y);
		}
		if (iso->w) {
			FreeM (iso->w);
		}
		if (iso->h) {
			int i;
			for (i=0; i<iso->n_filters; i++) {
				FreeM (iso->h[i].p);
			}
			FreeM (iso->h);
		}
		FreeM (iso);
		return 0;
	}
	return -1;
}


//==========================================================
// 
//----------------------------------------------------------
void ISOLearner_Input (ISOLearner* iso, real* x) 
{
	int i;
	real y = 0.0f;
	for (i=0; i<iso->n_filters; i++) {
		y += iso->h[i].f(&iso->h[i], x, iso->dt);
	}
	iso->y[0] = y;
}

//==========================================================
// 
//----------------------------------------------------------
int NewISOFilter (ISOFilter* h, int n_inputs)
{
	SpringMass* mass;
	h->f = &ISOFilterInput;
	h->u = 0.0f;

	if ((mass = AllocM (SpringMass, 1))==NULL) {
		Serror ("Could not create filter\n");
		return -1;
	}

	mass->a = 0.0f;
	mass->u = 0.0f;
	mass->x = 0.0f;
	mass->d = drand48();
	mass->k = drand48()*10.0f;
	h->p = (real*) mass;
  return 0;
}



//==========================================================
// 
//----------------------------------------------------------
real ISOFilterInput (struct ISOFilter_* h, real* x, real dt)
{
	SpringMass* s = (SpringMass*) h->p;
	s->a = *x - (s->k * s->x) - (s->d * s->u);
	s->u += s->a * dt;
	s->x += s->u * dt;
	return s->x;
}
