/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
const char* LINPRED_VERSION =  "VER: $Id: LinearPredictor_test.c,v 1.15 2003/08/28 00:09:04 olethros Exp $";
const char* VERSION_INFO = "LinearPredictor 0.1\n\
Copyright (C) 2002 Dimitrakakis Christos\n\
LinearPredictor comes with no WARRANTY,\n\
to the extent permitted by law.\n\
You may redistribute copies of LinearPredictor\n\
under the terms of the GNU General Public License.\n\
For more information about these matters\n\
see the files named COPYTING.\n";

const char* help = "\n\
  LinearPredictor_test <options>\n\n\
  --ascii         \t Files are in ASCII format\n\
  --16bit         \t Files are in 16bit binary format\n\
  --be            \t Assume data is big-endian (default: little-endian)\n\
  --unsigned      \t Assume data is unsigned (default: signed 2s-complement)\n\
  --concat        \t Data is concatenated (default: interleaved)\n\
  --classes       \t Number of classes to be recognized\n\
  --dimensions    \t Dimensionality of input\n\
  --frame         \t Frame-size to look at when predicting\n\
  --epochs        \t Number of epochs to train for\n\
  --lr            \t Set learning rate\n\
  --decay         \t Set learning rate decay (default: 1.0)\n\
  --train         \t Training file\n\
  --test          \t Test file\n\
  --output        \t Results output\n\
  --help       -h \t Show this help\n\n\
  --version    -v \t Show version information and exit\n\n\
Report bugs to Christos Dimitrakakis <dimitrak@idiap.ch>\n";


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <getopt.h>
#include "debug.h"
#include "SampleBuffer.h"
#include "datafile.h"
#include "psr.h"
#include "LinearPredictor.h"
#include "string_utils.h"
#include "LinkedList.h"


#undef DEBUG_ARGUMENTS

int main(int argc, char* argv[])                          
{
  int m, n, i, j, c;

  /* Options */
  int n_preds = 0;
  int n_examples = 0;
  int n_test = 0;
  int n_dim = 0;
  int pred_depth = 0;
  long cutoff;
  real learning_rate = 0.01f;
  real learning_rate_decay = 0.999f;
  bool do_decay = false;
  int epochs = 4096;
  FILE* msgport = NULL;
  char* foutname = NULL;
  char* foutname_test = NULL;
  FILE* outfile = NULL;
  FILE* outfile_test = NULL;
  int n_classes = 0;
  char* train_list = NULL;
  char* test_list = NULL;
  long n_read_type = UNKNOWN_FILETYPE;
  int sampling_rate = 22050;
  int window_period = 440;
  char* torch_train_name = NULL;
  char* torch_test_name = NULL;
  FILE* torch_train = NULL;
  FILE* torch_test = NULL;
  int n_torch_examples = 0;
  long total_errors = 0;
  long total_tests = 0;


  /* File pairs (inputs and targets) */
  FilePair train_pairs;
  FilePair test_pairs;

  /* Dynamically allocated */
  LinearPredictor** predictor = NULL;
  SAMPLE_BUFFER** train_buffer = NULL;
  SAMPLE_BUFFER** test_buffer = NULL;
  SAMPLE_BUFFER** train_target = NULL;
  SAMPLE_BUFFER** test_target = NULL;
  SAMPLE_BUFFER_LIST* train_buffer_list = NULL;
  SAMPLE_BUFFER_LIST* test_buffer_list = NULL;
  SAMPLE_BUFFER_LIST* train_target_list = NULL;
  SAMPLE_BUFFER_LIST* test_target_list = NULL;


  real* error_wave = NULL;

  msgport = stderr;



  //---------------------------- OPTIONS -----------------------------------//

  /* Set up default options */
  n_read_type = BIT16_FILETYPE;
  
  /* Read other options */
  while (1) {
    int option_index = 0;
    static struct option long_options[] = {
      {"ascii", no_argument, 0, 0}, //0
      {"16bit", no_argument, 0, 0}, //1
      {"classes", required_argument, 0, 0}, //2
      {"dimensions", required_argument, 0, 0}, //3
      {"frame", required_argument, 0, 0}, //4
      {"train", required_argument, 0, 0}, //5
      {"test", required_argument, 0, 0}, //6
      {"output", required_argument, 0, 0}, //7
      {"help", no_argument, 0, 'h'}, //8
      {"epochs", required_argument, 0, 0}, //9
      {"lr", required_argument, 0, 0}, //10
      {"decay", required_argument, 0, 0}, //11
      {"version", no_argument, 0, 'v'}, //12
      {"window", required_argument, 0, 0}, //13
      {"sampling_rate", required_argument, 0, 0}, //14
      {0, 0, 0, 0}
    };

    c = getopt_long (argc, argv, "vh",
                     long_options, &option_index);
    if (c == -1) {
      break;
    }
    switch (c) {
    case 0: /* This is the case for long arguments */
#ifdef DEBUG_ARGUMENTS
      printf ("option %s", long_options[option_index].name);
      if (optarg)
        printf (" with arg %s", optarg);
      printf ("\n");
#endif
      switch(option_index) {
      case 0: n_read_type = FLOAT_FILETYPE; break;
      case 1: n_read_type = BIT16_FILETYPE; break;
      case 2: n_classes = atoi (optarg); break;
      case 3: n_dim = atoi (optarg); break;
      case 4: pred_depth = atoi (optarg); break;
      case 5: train_list = optarg; break;
      case 6: test_list = optarg; break;
      case 7: foutname = optarg; break;
      case 9: epochs = atoi (optarg); break;
      case 10: learning_rate = atof (optarg); break;
      case 11:
        do_decay = true;
        if (optarg) {
          learning_rate_decay = atof (optarg);
        }else{
          learning_rate_decay = 0.999;
        }
        break;
      case 13:
        window_period = atoi (optarg);
        break;
      case 14:
        sampling_rate = atoi (optarg);
        break;
      default :
        Swarning ("This should never happen\n");
      }
      break;
      

    case 'h':
      printf("%s",help);
      exit(0);
      break;

    case '?':
      printf ("Ambiguous match\n");
      break;

    case ':':
      printf ("Missing parameter\n");
      break;

    default:
      printf ("?? getopt returned character code 0%o ??\n", c);
    }
  }
   
  if (optind < argc) {
    printf ("Stray arguments : ");
    while (optind < argc)
      printf ("%s ", argv[optind++]);
    printf ("\n");
    exit(-1);
  }

  printf("%s\n",LINPRED_VERSION);
  fprintf (msgport, " Classes: %d\n Dimensions: %d\n Depth: %d\n Trainfile: %s\n Testfile: %s\n Output: %s\n Learning Rate: %f\n Epochs: %d\n",n_classes, n_dim, pred_depth, train_list, test_list, foutname, learning_rate, epochs);
  if (do_decay) 
    fprintf (msgport, " Decay: %f\n", learning_rate_decay);



  //------------------------ CHECK ARGUMENT VALIDITY -----------------------//
  if (pred_depth<=0) {
    Serror("Frame size prediction depth should be larger than zero!\n");
    exit(-1);
  }

  if (n_classes<2) {
    Serror("You need to specify at least 2 classes!\n");
    exit(-1);
  }

  if (n_dim<1) {
    Serror("You need to specify at least 1 dimension!\n");
    exit(-1);
  }

  if (foutname==NULL) {
    Swarning ("No name specified for output. Using name 'output'\n");
    foutname = "output";
  }
  if (!strcmp(foutname,"")) {
    Serror ("Output filename is an empty string! Check your options\n");
    exit (-1);
  }
  
  if (window_period<1) {
    Serror ("Invalid window period\n");
    exit (-1);
  }

  if (sampling_rate<1) {
    Serror ("Invalid sampling rate\n");
    exit (-1);
  }

  foutname_test = strConcat (2, foutname, "_test");
  torch_train_name = strConcat (2, foutname, "_torch_train");
  torch_test_name = strConcat (2, foutname, "_torch_test");

  /* OK - check out the lists */
  if (readFileList(train_list, &train_pairs)) {
    Serror("Could not read training list\n");
    exit (-1);
  }
  if (readFileList(test_list, &test_pairs)) {
    Serror("Could not read testing list\n");
    exit (-1);
  }

  n_preds = n_classes;
  n_examples = train_pairs.n_files;
  n_test = test_pairs.n_files;
  cutoff = (pred_depth)*(n_dim);

  fprintf (msgport, "\n Output window period: %d samples (%fms @ %dHz)\n\n",
           window_period, 1000.0*((float) window_period)/((float) sampling_rate), sampling_rate);

  //---------------------------- SETUP -------------------------------------//
  /* Allocate and read buffers */
  fprintf (msgport, "Setting up\n");
  predictor = AllocM (LinearPredictor*, n_preds);
  error_wave = AllocM (real, n_preds);


  switch (n_read_type) {
  case FLOAT_FILETYPE:
    train_buffer_list =
      LoadASCIIBufferList (train_pairs.input_files, train_pairs.n_files);
    test_buffer_list =
      LoadASCIIBufferList (test_pairs.input_files, test_pairs.n_files);
    train_buffer = train_buffer_list->s;
    test_buffer = test_buffer_list->s;
    break;
  case BIT16_FILETYPE:
    train_buffer_list =
      LoadBufferList (train_pairs.input_files, train_pairs.n_files);
    test_buffer_list =
      LoadBufferList (test_pairs.input_files, test_pairs.n_files);
    train_buffer = train_buffer_list->s;
    test_buffer = test_buffer_list->s;
    break;
  default:
    Serror ("I don't know how to load this data. Sorry.\n");
    exit (-1);
  }

  train_target_list =
    LoadASCIIBufferList (train_pairs.target_files, train_pairs.n_files);
  test_target_list =
    LoadASCIIBufferList (test_pairs.target_files, test_pairs.n_files);
  train_target = train_target_list->s;
  test_target = test_target_list->s;

  if (WriteFile (foutname, outfile)) {
    fprintf (msgport, "Writing output to %s\n",foutname);
  }else{
    Serror ("Failed to write output!");
    return -1;
  }
  
  if (WriteFile (foutname_test, outfile_test)) {
    fprintf (msgport, "Writing test output to %s\n",foutname_test);
  }else{
    Serror ("Failed to write output!");
    return -1;
  }

  if (WriteFile (torch_train_name, torch_train)) {
    fprintf (msgport, "Writing torch test output to %s\n", torch_train_name);
  }else{
    Serror ("Failed to write output!");
    return -1;
  }

  
  if (WriteFile (torch_test_name, torch_test)) {
    fprintf (msgport, "Writing torch test output to %s\n",torch_test_name);
  }else{
    Serror ("Failed to write output!");
    return -1;
  }

  





  fprintf (msgport, "Created %d sample buffers\n", n_examples);

  /* Create predictors */
  for (m=0; m<n_preds; m++) {
    predictor[m] = NewLinearPredictor(pred_depth, n_dim);
    LinearPredictor_Reset(predictor[m]);
    LinearPredictor_SetLearningRate(predictor[m], learning_rate);
  }
  fprintf (msgport, "Created %d predictors\n",n_preds);
  if (do_decay) {
    fprintf (msgport, "Learning rate : [%f, %f]\n", learning_rate, 
             learning_rate*pow (learning_rate_decay, epochs));
  }


  /* Allocate examples to each predictor */
  /* (this should behave differently depending on how targets are specified) */

  for (m=0; m<train_pairs.n_files; m++) {
    if (train_target[m]->length != 1) {
      Serror ("Buffer %d has length %ld! All target lengths should be one. Don't know what to do!\n", m, train_target[m]->length);
      exit (-1);
    }

    i = (int) (*(train_target[m]->x));
    if ( (i<0) || (i>n_classes)) {
      Serror ("Label exceeds number of classes. Aborting.\n");
      exit (-1);
    }
  }
  

  //----------------------------- TRAIN -----------------------------------//
  
  for (m=0; m<epochs; m++) {
    fprintf (stderr, "%d/%d\n", m, epochs);
    for (n=0; n<train_pairs.n_files; n++) {
      real* x;
      real error = 0.0f;
      x = train_buffer[n]->x;
      i = ((int) (*(train_target[n]->x))) - 1;
      assert(predictor[i]);
      assert(x);
      for (j=0; j<(train_buffer[n]->length - cutoff)/n_dim; j++) {
        LinearPredictor_Train (predictor[i], x);
        error += LinearPredictor_GetError (predictor[i]);
        x += n_dim;
      }
      fprintf (outfile, "%f ",error);
    }
    fprintf(outfile, "\n");
  }

  //----------------------------- TEST ------------------------------------//
  if (error_wave) {
    for (i=0; i<n_preds; i++) {
      error_wave[i] = 0.0f;
    }
  } else {
    Serror ("Oops\n");
    exit (-1);
  }

  fprintf (msgport, "Testing....\n");
  total_errors = 0;
  total_tests = 0;
 

  //----------------------------- MEASURE TRAIN -------------------------------//
  n_torch_examples = 0;
  for (n=0; n<train_pairs.n_files; n++) {
      n_torch_examples += ((train_buffer[n]->length-cutoff)/n_dim)/(window_period+1);
      fprintf(msgport, "%ld windows in train file %d\n", ((train_buffer[n]->length-cutoff)/n_dim)/(window_period+1), n);
  }
  fprintf (torch_train, "%d %d\n", n_torch_examples, n_preds+1);
  for (i=0; i<n_preds; i++) {
    error_wave[i] = 0;
  }

  for (n=0; n<train_pairs.n_files; n++) {
    real* x;
    int target;
    real mini;
    real min;
    int count = window_period;

    x = train_buffer[n]->x;
    for (j=0; j<(train_buffer[n]->length - cutoff)/n_dim; j++) {
      target = ((int) (*(train_target[n]->x))) - 1;
      for (i=0; i<n_preds; i++) {
        error_wave[i] += LinearPredictor_Test (predictor[i], x);
        if (count==0) {
#ifdef SHOW_THE_OUTPUTS_INSTEAD
          real* monitor = LinearPredictor_GetOutput (predictor[i]);
          for (int k=0; k<n_dim; k++) {
            fprintf (outfile, "%f ", monitor[k]);
          }
#else
          fprintf (outfile, "%f ",error_wave[i]);
          fprintf (torch_train, "%f ",error_wave[i]);
#endif
        }
      }
      if (count==0) {
        fprintf (torch_train, "%d\n",target);
        min = error_wave[0];
        mini = 0;
        for (i=0; i<n_preds; i++) {
          if (min>error_wave[i]) {
            min = error_wave[i];
            mini = i;
          }
          error_wave[i] = 0.0f;
        }
        if (mini!=target) {
          total_errors ++;
        }
        total_tests++;
      }
      x += n_dim;
      if (count==0) {
        fprintf (outfile, "\n");
      }
      if (count==0) count = window_period;
      count --;
    }
  }

  printf ("Test error : %f\n", ((float) total_errors)/((float)total_tests));

  
  
  //----------------------------- MEASURE TEST -------------------------------//
  total_errors = 0;
  total_tests = 0;

  n_torch_examples = 0;
  for (n=0; n<test_pairs.n_files; n++) {
    n_torch_examples += ((test_buffer[n]->length-cutoff)/n_dim)/(window_period+1);
    fprintf(msgport, "%ld windows in test file %d\n", ((test_buffer[n]->length-cutoff)/n_dim)/(window_period+1), n);
  }
  fprintf (torch_test, "%d %d\n", n_torch_examples, n_preds+1);
  for (i=0; i<n_preds; i++) {
    error_wave[i] = 0;
  }





  for (n=0; n<test_pairs.n_files; n++) {
    real* x;
    int target;
    real mini;
    real min;
    int count = window_period;

    x = test_buffer[n]->x;
    for (j=0; j<(test_buffer[n]->length - cutoff)/n_dim; j++) {
      if (count==0) {
        for (int k=0; k<n_dim; k++) {
          fprintf (outfile_test, "%f ", x[k]);
        }
      }
      target = ((int) (*(test_target[n]->x))) - 1;
      for (i=0; i<n_preds; i++) {
        error_wave[i] += LinearPredictor_Test (predictor[i], x);
        if (count==0) {
#ifdef SHOW_THE_OUTPUTS_INSTEAD
          real* monitor = LinearPredictor_GetOutput (predictor[i]);
          for (int k=0; k<n_dim; k++) {
            fprintf (outfile_test, "%f ", monitor[k]);
          }
#else
          fprintf (outfile_test, "%f ",error_wave[i]);
          fprintf (torch_test, "%f ",error_wave[i]);
#endif
        }
      }
      if (count==0) {
        fprintf (torch_test,"%d\n",target);
        min = error_wave[0];
        mini = 0;
        for (i=0; i<n_preds; i++) {
          if (min>error_wave[i]) {
            min = error_wave[i];
            mini = i;
          }
          error_wave[i] = 0.0f;
        }
        if (mini!=target) {
          total_errors ++;
        }
        total_tests++;
      }
      x += n_dim;
      if (count==0) {
        fprintf (outfile_test, "\n");
      }
      if (count==0) count = window_period;
      count --;
    }
  }


  printf ("Test error : %f\n", ((float) total_errors)/((float)total_tests));



  //---------------------------- CLEANUP ----------------------------------//
  
  fprintf (msgport, "Cleaning up\n");
  for (m=0; m<n_preds; m++) {
    DeleteLinearPredictor(predictor[m]);
  }
  FreeM (predictor);
  FreeM (error_wave);

  FreeBufferList (train_buffer_list);
  train_buffer = NULL;

  FreeBufferList (test_buffer_list);
  test_buffer = NULL;

  FreeBufferList (train_target_list);
  train_target = NULL;

  FreeBufferList (test_target_list);
  test_target = NULL;

  
  FreeM (torch_train_name);
  FreeM (torch_test_name);

  FreeM (foutname_test);
  CloseFile (outfile);
  CloseFile (outfile_test);
  CloseFile (torch_train);
  CloseFile (torch_test);

  fprintf (msgport, "Exiting\n");
  return 0;
}

 

