#include "LinkedList.h"


BUFFER* NewBuffer(long length)
{
  BUFFER* b = NULL;
  b = AllocM(BUFFER, 1);
  //  assert(b);
  if (!b) {
	Serror("ERROR");
  }
  b->data = malloc(length);
  if (!b->data) {
	Serror("could not allocate space for data\n");
  }
  b->len = length;

  return b;
}

int DeleteBuffer(BUFFER* Buffer)
{
  assert(Buffer);
#ifdef DEBUG_LINKEDLIST
  printf("deleting item contents : %s\n",(char*) Buffer->data);
#endif
  FreeM(Buffer->data);
  Buffer->len = 0;
  FreeM(Buffer);
  return 0;
}


LISTNODE* AddItem(LINKEDLIST* list, BUFFER* buffer)
{
  assert (buffer);

  if (list->last) {
	assert(list->last->data);
	list->last->next = NewListItem();
	list->last->next->ID = list->last->ID + 1;
	list->last = list->last->next;
  } else {
	list->last = NewListItem();
	list->first = list->last;
	list->last->ID = 0;

  }


  list->last->data = buffer;
  list->curr = list->last;
  
  list->items++;
#ifdef DEBUG_LINKEDLIST
  printf("Added item - %ld total\n",list->items);
#endif

  return list->curr;
  
}

LISTNODE* NewListItem(void) {
  LISTNODE* node = NULL;
  node = AllocM (LISTNODE, 1);
  node->data = NULL;
  node->next = NULL;
  return node;
}

int DeleteItem(LISTNODE* node)
{
  assert(node);
  DeleteBuffer(node->data);
  node->data = NULL;
  node->next = NULL;
  node->ID = 0;
  FreeM(node);

  return 0;
}


LINKEDLIST* NewList(void)
{
  LINKEDLIST* list = NULL;
  list = AllocM(LINKEDLIST, 1);
  list->first = NULL;
  list->curr = NULL;
  list->last = NULL;
  list->items = 0;
  
  return list;
}

int DeleteList(LINKEDLIST* list)
{
  LISTNODE* item = NULL;
  LISTNODE* nextitem = NULL;

  assert(list);
  item = list->first;
  assert(item);

  do {
	nextitem = item->next;
	DeleteItem(item);
	item = nextitem;
	list->items--;
  } while (nextitem);

#ifdef DEBUG_LINKEDLIST
  printf("List must be clear - %ld items remaining\n", list->items);
#endif

  if (list->items) {
	Serror("List is not clear, while it should be - %ld items remaining\n", list->items);
  }

  FreeM(list);

  return 0;
}


/* 
   GetStringMatchingListItem
   
   Scans through a linked list until a particular item is found.
   
   - USES: strcmp(), strcpy()
   (Note: Take particular care when suing this function, as it assumes that all buffers pointed to by the list nodes contain valid null-terminated strings! - Points where potential conflicts might arise are indicated in the (char *) casts)

 */

LISTNODE* GetStringMatchingListItem(LINKEDLIST* list, char* token) {
  LISTNODE* node = NULL;
  LISTNODE* MatchingNode = NULL;
  assert(list);
  node = list->first;
 
  while(node) {
	/* WARNING: This code assumes that the data held in the buffer is a null-terminated string! */
	if (strcmp(token, (char*) node->data->data)) {
	  node = node->next;
	} else {
	  MatchingNode = node;
	  node = NULL;
	}
  }

  if (!MatchingNode) {
	BUFFER* b = NULL;
	long l = strlen(token)+1;
	b = NewBuffer(l);
	strcpy((char*) b->data, token);
	MatchingNode = AddItem(list, b);
  }

  return MatchingNode;
 
}

int Length (LINKEDLIST* list)
{
  return list->items;
}
