/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* VER: $Id: List.h,v 1.6 2003/09/06 17:26:18 olethros Exp olethros $ */

#ifndef LIST_H
#define LIST_H

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>


typedef struct ListItem {
  void* obj;
	void (*free_obj) (void* obj);
  struct ListItem* prev;
  struct ListItem* next;
} LISTITEM;


/// A very simple list structure
typedef struct List {
  LISTITEM* curr;
  LISTITEM* head;
  LISTITEM* tail;
  int n;
/// Method by which to search objects
  LISTITEM* (*retrieve) (struct List* list, void* ptr);
} LIST;


/// Create a new list
LIST* List(void);
/// Get the size of the list
int ListSize(LIST* list);
/// Append an item to the list
LISTITEM* ListAppend(LIST* list, void* p);
/// Append an item to the list with free hook
LISTITEM* ListAppend(LIST* list, void* p, void (*free_obj) (void* obj));
/// Move to the first list item
LISTITEM* FirstListItem(LIST* list);
/// Move to the first list item
LISTITEM* LastListItem(LIST* list);
/// Advance one item
LISTITEM* NextListItem(LIST* list);
/// Remove the topmost item of the list (also frees obj memory)
int PopItem(LIST* list);
/// Clear the list
int ClearList(LIST* list);
/// Finds the LISTITEM pointer corresponding to the data
LISTITEM* FindItem (LIST* list, void* ptr);
/// Get the nth item of the list
LISTITEM* GetItem (LIST* list, int n);

LISTITEM* ListItem(void* ptr, void (*free_obj) (void* obj));
LISTITEM* GetNextItem(LISTITEM* ptr);
LISTITEM* GetPrevItem(LISTITEM* ptr);
LISTITEM* LinkNext(LISTITEM* src, void* ptr, void (*free_obj) (void* obj));
LISTITEM* LinkPrev(LISTITEM* src, void* ptr, void (*free_obj) (void* obj));
int FreeListItem(LIST* list, LISTITEM* ptr);
int RemoveListItem(LIST* list, LISTITEM* ptr);

LISTITEM* ListLinearSearchRetrieve (struct List* list, void* ptr);

#endif
