/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/*
  This program creates two files.
  A test and a training file.
*/
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <time.h>
#include "datafile.h"
#include "LinkedList.h"
#include "MarkovChain.h"


#define N_OUT_FILES 2 /* Number of output files */
#define N_CLASSES 2 /* Number of classes */
#define MC_STATES 4 /* Number of states in the Markov Chain */
#define MC_ORDER 1 /* Order of the Markov Chain */

int main (int argc, char* argv[])
{
  char* file_name[N_OUT_FILES];
  FILE* file_stream[N_OUT_FILES];
  MarkovChain* chain[N_CLASSES];
  int i, files;
  int seq;
  int error_flag = 0;

  srand48 (time (NULL));

  if (argc<=N_OUT_FILES) {
    fprintf (stderr, "I need %d filenames!\n", N_OUT_FILES);
    return -1;
  }

  for (i=0; i<N_OUT_FILES; i++) {
    file_name[i] = argv[i+1];
    if (WriteFile(file_name[i], file_stream[i])) {
      printf ("Opened %s for output\n", file_name[i]);
    }
  }

  printf ("Creating two markov chains for the two classes\n");
  for (i=0; i<N_CLASSES; i++) {
    if ((chain[i] = NewMarkovChain (MC_STATES, MC_ORDER))==NULL) {
      error_flag++;
    }
  }

  if (error_flag) {
    goto quick_exit;
  }
  

  /* Here, we have no data to train the chains of course, we just set up their transition probabilities, god help us */
  for (i=0; i<N_CLASSES; i++) {
    int j, k;
    int n_states;
    float* trans;
    trans = chain[i]->transitions;
    MarkovChainReset (chain[i]);
    n_states = chain[i]->tot_states;
    for (j=0; j<chain[i]->tot_states; j++) {
      for (k=0; k<chain[i]->n_states; k++) {
	trans[j+k*n_states] = drand48 ();
      }
    }
    MarkovChainNormalize (chain[i]);
    MarkovChainShowTransitions (chain[i]);
  }

  /* Now try and make some examples from the states */

  for (files=0; files<N_OUT_FILES; files++) {
    FILE* output_stream = file_stream[files];
    int j;
    seq = 0;
    for (j=0; j<100; j++) {
      for (i=0; i<N_CLASSES; i++) {
	int n_examples = rand()%8;
	int n;
	MarkovChainReset(chain[i]);
	for (n=0; n<n_examples; n++, seq++) {
	  int m;
	  int x;
	  x = MarkovChainGenerate(chain[i]);
	  MarkovChainPushState (chain[i], x);
	  //printf ("CHAIN %d!", i);
	  //    MarkovChainShowTransitions (chain[i]);
	  for (m=0; m<MC_STATES; m++) {
	    if (m==x) {
	      fprintf (output_stream, "1 ");
	    } else {
	      fprintf (output_stream, "0 ");
	    }
	  }
	  fprintf (output_stream, "%d %d\n", i, seq);
	}
      }
    }
  }		    

 quick_exit:
  if (error_flag) {
    fprintf (stderr, "%d errors detected - aborting\n", error_flag);
  }
  printf ("Cleaning up\n");
  for (i=0; i<N_CLASSES; i++) {
    DeleteMarkovChain(chain[i]);
  }

  for (i=0; i<N_OUT_FILES; i++) {
    CloseFile(file_stream[i]);
  }

  printf ("Bye!\n");
  return error_flag;
}
