/* -*- Mode: C++; -*- */
/* VER: $Id: Predictor.c,v 1.8 2003/09/05 01:08:43 olethros Exp $ */
#include "Predictor.h"


Predictor* NewPredictor (int n_inputs, int d, int T, int look_ahead)
{
	Predictor* p = AllocM (Predictor, 1);
	p->d = d;
	p->T = T;
	p->look_ahead = look_ahead;
	p->n_inputs = n_inputs;
	p->x = NULL;
	p->y = NULL;
	p->windowing_function = NULL;
	return p;
}

Predictor* NewANNPredictor (int n_inputs, int d, int T, int look_ahead)
{
	int l, ly;
	Predictor* p = NewPredictor (n_inputs, d, T, look_ahead);	
	l =  p->d * p->n_inputs;
	ly =  (p->look_ahead + p->T) * p->n_inputs;
	p->ann = NewANN (l, ly);
	p->x = AllocM (real, l);
	p->y = AllocM (real, ly);
	return p;
}

Predictor* NewHCNNPredictor (int n_inputs, int d, int T, int look_ahead)
{
	int l, ly;
	Predictor* p = NewPredictor (n_inputs, d, T, look_ahead);
	l = 1 + p->d * p->n_inputs;
	ly =  (p->look_ahead + p->T) * p->n_inputs;
	p->ann = NewANN (l, ly);
	p->x = AllocM (real, l);
	p->y = AllocM (real, ly);
	return p;
}


int DeletePredictor (Predictor* pred)
{
	int r = DeleteANN (pred->ann);
	if (pred->x) {
		FreeM (pred->x);
	} else {
		Swarning ("there should be a pred->x\n");
	}
	if (pred->y) {
		FreeM (pred->y);
	} else {
		Swarning ("there should be a pred->y\n");
	}
	FreeM (pred);
	return r;
}
int Predictor_AddHiddenLayer (Predictor* pred, int n_units)
{
	return ANN_AddHiddenLayer (pred->ann, n_units);
}

int Predictor_AddRBFHiddenLayer (Predictor* pred, int n_units)
{
	return ANN_AddRBFHiddenLayer (pred->ann, n_units);
}

void Predictor_SetLearningRate (Predictor* pred, real a)
{
	ANN_SetLearningRate (pred->ann, a);
}
	
void Predictor_SetBatchMode (Predictor* pred, bool batch)
{
	ANN_SetBatchMode (pred->ann, batch);
}
	
void Predictor_BatchAdapt(Predictor* pred)
{
	ANN_BatchAdapt (pred->ann);
}

void Predictor_SetOutputsToTanH (Predictor* pred)
{
	ANN_SetOutputsToTanH (pred->ann);
}

int Predictor_Init (Predictor* pred)
{
	int r = 0;
	r = ANN_Init (pred->ann);
	ANN_SetOutputsToLinear (pred->ann); //we are doing regression
	return r;
}

/* Functionality Interface */
real Predictor_Input(Predictor* predictor, real* x)
{
	Predictor_CompactInputs (predictor, x);
	return ANN_Input (predictor->ann, predictor->x);
}

real Predictor_HCNNInput(Predictor* predictor, real* x, real c)
{
	Predictor_HCNNCompactInputs (predictor, x, c);
	return ANN_Input (predictor->ann, predictor->x);
}

real Predictor_Train(Predictor* predictor, real* x)
{
	Predictor_CompactInputs (predictor, x);
	return ANN_Train (predictor->ann, predictor->x, predictor->y);
}


real Predictor_HCNNTrain(Predictor* predictor, real* x, real c)
{
	Predictor_HCNNCompactInputs (predictor, x, c);
	return ANN_Train (predictor->ann, predictor->x, predictor->y);

}

real Predictor_Test(Predictor* predictor, real* x)
{
	Predictor_CompactInputs (predictor, x);
	return ANN_Test (predictor->ann, predictor->x, predictor->y);
}

real* Predictor_GetOutput(Predictor* predictor)
{
	return ANN_GetOutput (predictor->ann);
}
real Predictor_GetError(Predictor* predictor)
{
	return ANN_GetError (predictor->ann);
}

void Predictor_CompactInputs(Predictor* predictor, real* x)
{
	long i;
	long lx = predictor->d * predictor->n_inputs;
	long ly = (predictor->look_ahead + predictor->T) * predictor->n_inputs;
	real* x_buff = predictor->x;
	real* y_buff = predictor->y;
	real* x_data = x;
	int T = predictor->T;
	SampleWindow* wf = predictor->windowing_function;
	
	if (predictor->windowing_function) {
		SAMPLE_BUFFER buf= {lx, x_data, 0};
		for (i=0; i<lx; i++) {
			const SAMPLE_BUFFER* ret = wf->Cut(&buf, i*T, 0);
			real v = 0.0f;
			for (int j=0; j<T; j++) {
				v += ret->x[j];
			}
			x_buff [i] = v / T;
			
		}
	} else {
		for (i=0; i<lx; i++) {
			x_buff[i] = x_data[i*T];
		}
	}

	x_data = &x[T*(lx-1)+1];
	for (i=0; i<ly; i++) {
		*y_buff++ = *x_data++;
	}
} 

void Predictor_HCNNCompactInputs(Predictor* predictor, real* x, real c) 
{
	int l = predictor->d * predictor->n_inputs;
	Predictor_CompactInputs (predictor, x);
	predictor->x[l] = c;
}

void Predictor_SetWindowingFunction (Predictor* predictor, SampleWindow* window)
{ 
	predictor->windowing_function = window;
}
