/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
const char* PREDCLASS_VERSION = "VER: $Id: PredictorClassifier.c,v 1.4 2003/08/28 00:09:04 olethros Exp $";
const char* VERSION_INFO = "PredictorClassifier 0.1\n\
Copyright (C) 2003 Christos Dimitrakakis\n\
PredictorClassifier comes with no WARRANTY,\n\
to the extent permitted by law.\n\
You may redistribute copies of PredictorClassifier\n\
under the terms of the GNU General Public License.\n\
For more information about these matters\n\
see the files named COPYTING.\n";

const char* help = "\n\
  TestPredictor <options>\n\n\
  --inputs        \t Dimensionality of input\n\
  --delay         \t Delay of Data [1]\n\
  --dimension     \t Embedding dimension [32]\n\
  --ahead         \t Look-ahead [0]\n\
  --epochs        \t Number of epochs to train for [100]\n\
  --nhu           \t Number of hidden units [25]\n\
  --nhu2          \t Number of hidden units in the 2nd hidden layer [0]\n\
  --lr            \t Set learning rate [0.1]\n\
  --decay         \t Set learning rate decay [1.0]\n\
  --train         \t Training file\n\
  --average       \t Average for prediction\n\
  --batch         \t Use batch mode for training\n\
  --help       -h \t Show this help\n\n\
  --version    -v \t Show version information and exit\n\n\
Report bugs to Christos Dimitrakakis <dimitrak@idiap.ch>\n";


#include <time.h>
#include <getopt.h>
#include <assert.h>
#include "Predictor.h"
#include "SampleBuffer.h"
#include "datafile.h"
#include "psr.h"


int main (int argc, char** argv)
{
	int n_iter = 25;//100;
	int window_size;
  int cutoff;
  int n_classes = 0;
  int n_inputs = 1;
  int n_hidden = 25;
  int n_hidden_b = 0;
  bool rbf_layer = false;
  bool tanh_outputs = false;
  bool average = false;
  int delay = 1;
  int look_ahead = 0;
  int embedding_dimension = 32;
  real learning_rate = 0.01f;
  real learning_rate_decay = 1.0;//0.999f;
  int prior = 0;
  bool batch_mode = false;
  char* train_fname = NULL;
  char* test_fname = NULL;
  real stat_rms_train = 0.0f;
  real stat_rms_test = 0.0f;
  real stat_class_err = 0.0f;

  /* File pairs (inputs and targets) */
  FilePair train_pairs;
  FilePair test_pairs;

  SAMPLE_BUFFER_LIST* train_data = NULL;
  SAMPLE_BUFFER_LIST* test_data = NULL;
  SAMPLE_BUFFER_LIST* train_target_list = NULL;
  SAMPLE_BUFFER_LIST* test_target_list = NULL;
  SAMPLE_BUFFER** train_target = NULL;
  SAMPLE_BUFFER** test_target = NULL;
  Predictor** p;  
  int n_predictors = 0;




const char* help = "\n\
  TestPredictor <options>\n\n\
  --prior         \t Also use a general predictor\n\
  --inputs        \t Dimensionality of input\n\
  --classes       \t Number of Classes\n\
  --delay      -T \t Delay of Data [1]\n\
  --dimension  -d \t Embedding dimension [32]\n\
  --ahead      -a \t Look-ahead [0]\n\
  --epochs     -N \t Number of epochs to train for [100]\n\
  --rbf           \t Use RBFs for the first hidden layer\n\
  --tanh          \t Use Hyperbolic Tan at the output\n\
  --nhu           \t Number of hidden units [25]\n\
  --nhu2          \t Number of hidden units in the 2nd hidden layer [0]\n\
  --lr            \t Set learning rate [0.1]\n\
  --decay         \t Set learning rate decay [1.0]\n\
  --average       \t Average over outputs for prediction\n\
  --train         \t Training file\n\
  --test          \t Testing file\n\
  --help       -h \t Show this help\n\n\
  --version    -v \t Show version information and exit\n\n\
Report bugs to Christos Dimitrakakis <dimitrak@idiap.ch>\n";
  while (1) {
    int option_index = 0;
    static struct option long_options[] = {
      {"inputs", required_argument, 0, 0 }, //0
      {"delay", required_argument, 0, 'T'}, //1
      {"dimension", required_argument, 0, 'd'}, //2
      {"ahead", required_argument, 0, 'a'}, //3
      {"epochs", required_argument, 0, 'N'}, //4
      {"nhu", required_argument, 0, 0}, //5
      {"nhu2", required_argument, 0, 0}, //6
      {"lr", required_argument, 0, 0}, //7
      {"decay", required_argument, 0, 0}, //8
      {"train", required_argument, 0, 0}, //9
      {"rbf", no_argument, 0, 0}, //10
      {"tanh", no_argument, 0, 0}, //11
      {"average", no_argument, 0, 0}, //12
      {"version", no_argument, 0, 'v'}, //13
      {"help", no_argument, 0, 'h'}, //14
      {"test", required_argument, 0, 0}, //15
      {"classes", required_argument, 0, 0}, //16
      {"prior", no_argument, 0, 0},//17
      {"batch", no_argument, 0, 0},//18
      {0, 0, 0, 0}
    };

    int c = getopt_long (argc, argv, "i:d:a:N:T:vh",
                         long_options, &option_index);
    if (c == -1) {
      break;
    }
    switch (c) {
    case 0: /* This is the case for long arguments */
#ifdef DEBUG_ARGUMENTS
      printf ("option %s", long_options[option_index].name);
      if (optarg)
        printf (" with arg %s", optarg);
      printf ("\n");
#endif

      switch(option_index) {

      case 5: n_hidden = atoi (optarg); break;
      case 6: n_hidden_b = atoi (optarg); break;
      case 7: learning_rate = atof (optarg); break;
      case 8: learning_rate_decay = atof (optarg); break;
      case 9: train_fname = optarg; break;
      case 10: message ("Using RBFs\n"); rbf_layer = true; break;
      case 11:
        message ("Setting activation to tanh\n");
        tanh_outputs = true;
        break;
      case 12:
        message ("Averaging for prediction\n");
        average = true;
        //        break;
      case 13: printf ("Version\n"); break;
      case 14: printf ("HELP\n"); break;
      case 15: test_fname = optarg; break;
      case 16: n_classes = atoi (optarg); break;
      case 17: prior = 1; break;
      case 18: batch_mode = true; break;
      default :
        Swarning ("This should never happen\n");
      }
      break;
      
    case 'i': n_inputs = atoi (optarg); break;
    case 'd': embedding_dimension = atoi (optarg); break;
    case 'a': look_ahead = atoi (optarg); break;
    case 'N': n_iter = atoi (optarg); break;
    case 'T': delay = atoi (optarg); break;

    case 'v':
      printf ("%s", PREDCLASS_VERSION);
      break;
    case 'h':
      printf("%s",help);
      exit(0);
      break;
    case '?':
      printf ("Ambiguous match\n");
      break;
    case ':':
      printf ("Missing parameter\n");
      break;
    default:
      printf ("?? getopt returned character code 0%o ??\n", c);
    }
  }

  message ("%s", PREDCLASS_VERSION);
  message ("N_ITER:%d",  n_iter);
  message ("n_classes:%d", n_classes);
  message ("n_inputs:%d", n_inputs);
  message ("n_hidden:%d",  n_hidden);
  message ("n_hiddenb:%d",  n_hidden_b);
  message ("rbf_layer %d", rbf_layer);
  message ("tanh %d", tanh_outputs);
  message ("average %d",  average);
  message ("T %d", delay);
  message ("l %d", look_ahead);
  message ("d %d", embedding_dimension);
  message ("lr %f", learning_rate);
  message ("lrd %f", learning_rate_decay);
  message ("prior %d", prior);

  window_size = embedding_dimension * delay ;
  cutoff = window_size + look_ahead + delay;

  n_predictors = n_classes + prior;
  if (n_classes<=1) {
    Serror ("There must be at least two classes");
    exit (-1);
  }

  /*** Initialise predictors ***/
  p = AllocM (Predictor*, n_predictors);
  for (int i=0; i<n_predictors; i++) {
    p[i]=  NewANNPredictor (n_inputs, embedding_dimension, delay, look_ahead);
    if (n_hidden > 0) {
      if (rbf_layer) {
        Predictor_AddRBFHiddenLayer (p[i], n_hidden);
      } else {
        Predictor_AddHiddenLayer (p[i], n_hidden);
      }
      if (n_hidden_b > 0) {
        Predictor_AddHiddenLayer (p[i], n_hidden_b);
      }
    }
    Predictor_Init (p[i]);
    Predictor_SetLearningRate (p[i], learning_rate);
    Predictor_SetBatchMode (p[i], batch_mode);
    if (tanh_outputs) {
      Predictor_SetOutputsToTanH (p[i]);
    }
  }


  /*** Load Data ***/

  /* OK - check out the lists */
  if (readFileList(train_fname, &train_pairs)) {
    Serror("Could not read training list\n");
    exit (-1);
  }
  if (readFileList(test_fname, &test_pairs)) {
    Serror("Could not read testing list\n");
    exit (-1);
  }

  train_data = LoadBufferList (train_pairs.input_files, train_pairs.n_files);
  test_data = LoadBufferList (test_pairs.input_files, test_pairs.n_files);

  /* Check the targets */
  train_target_list =
    LoadASCIIBufferList (train_pairs.target_files, train_pairs.n_files);
  test_target_list =
    LoadASCIIBufferList (test_pairs.target_files, test_pairs.n_files);
  train_target = train_target_list->s;
  test_target = test_target_list->s;

  for (int m=0; m<train_pairs.n_files; m++) {
    if (train_target[m]->length != 1) {
      Serror ("Buffer %d has length %ld! All target lengths should be one. Don't know what to do!\n", m, train_target[m]->length);
      exit (-1);
    }

    int i = (int) (*(train_target[m]->x));
    if ( (i<0) || (i>=n_classes)) {
      Serror ("Label exceeds number of classes. Aborting.\n");
      exit (-1);
    }
  }

  /*** Training Loop ***/
  for (int big_iter=0; big_iter<4; big_iter++) {
  message ("Training for %d iterations", n_iter);
  if (prior) {
    message ("(First training the prior predictor)");

    for (int iter=0; iter<n_iter; iter++) {
      real sum = 0.0f;
      int total_length = 0;
      /* learn by one, or by delay */
      for (int j = 0; j<train_data->n; j++) {
        real* data = train_data->s[j]->x;
        int max_len = train_data->s[j]->length - cutoff;
        for (int i=0; i<max_len; i+=delay) {
          sum += Predictor_Train (p[n_classes], &data[i]);
        }
        total_length += max_len;
      }
      sum = sqrt (sum / ((real) total_length));
      if (iter == n_iter-1) {
        message ("rms: %f", sum);
        stat_rms_train = sum;
      }
      if (batch_mode) {
        Predictor_BatchAdapt (p[n_classes]);
      }
    }
  }
  

  message ("(Training classes)");
  for (int iter=0; iter<n_iter; iter++) {
    real sum = 0.0f;
    int total_length = 0;

    /* learn by one, or by delay */
    for (int j = 0; j<train_data->n; j++) {
      real* data = train_data->s[j]->x;
      int max_len = train_data->s[j]->length - cutoff;
      int label = ((int) (*(train_target[j]->x)));
      assert (( (label>=0) && (label<n_classes) ));
      SAMPLE_BUFFER* error_buffer = NULL;
      if (prior) {
        error_buffer = NewSampleBuffer (max_len+delay, CLEAR_SAMPLE_BUFFER);
        for (int i=0; i<max_len; i+=delay) {
          Predictor_Input (p[n_classes], &data[i]);
          //          real* y = Predictor_GetOutput (p[n_classes]);
          //          real* b = error_buffer->x;
          for (int t=0; t<delay; t++) {
            //            int index = window_size+i-p[n_classes]->T+t+1;
            if ((i+t)>=error_buffer->length) {
              Serror ("Buffer Length %ld, trying to write to %d+%d=%d (c:%d)\n", error_buffer->length, i,t,i+t,cutoff);
            }
            //            b[i+t] = data[index] - y[t];
          }
        }
        max_len = max_len - cutoff;
        data = error_buffer->x;
      }
      for (int i=0; i<max_len; i+=delay) {
        real e = Predictor_Train (p[label], &data[i]);
        sum += e;
      }
      total_length += max_len;
      if (error_buffer) {
        FreeSampleBuffer (error_buffer);
      }
    }
    sum = sqrt (sum / ((real) total_length));
    if (iter==(n_iter-1)) {
      message ("rms: %f", sum);
      stat_rms_test = sum;
    }
    if (batch_mode) {
      for (int label=0; label<n_classes; label++) {
        Predictor_BatchAdapt (p[label]);
      }
    }
  }


  /*** Testing ***/
  message ("Testing");
  int class_err = 0;
  for (int j=0; j<test_data->n; j++) {
    real* data = train_data->s[j]->x;
    int max_len = test_data->s[j]->length - cutoff;
    int label = ((int) (*(train_target[j]->x)));
    assert (( (label>=0) && (label<n_classes) ));
    SAMPLE_BUFFER* error_buffer = NULL;
    if (prior) {
      error_buffer = NewSampleBuffer (max_len+delay, CLEAR_SAMPLE_BUFFER);
      for (int i=0; i<max_len; i+=delay) {
        Predictor_Input (p[n_classes], &data[i]);
        real* y = Predictor_GetOutput (p[n_classes]);
        real* b = error_buffer->x;
        for (int t=0; t<delay; t++) {
          int index = window_size+i-p[n_classes]->T+t+1;
          if ((i+t)>=error_buffer->length) {
            Serror ("Buffer Length %ld, trying to write to %d+%d=%d (c:%d)\n", error_buffer->length, i,t,i+t,cutoff);
          }
          b[i+t] = data[index] - y[t];
        }
      }
      max_len = max_len - cutoff;
      data = error_buffer->x;
    }
    real* sum = AllocM(real, n_classes);
    real minerr;
    int minerr_i = 0;
    for (int cl=0; cl<n_classes; cl++) {
      sum[cl] = 0.0f;
      for (int i=0; i<max_len; i+=delay) {
        sum[cl] += Predictor_Train (p[cl], &data[i]);
      }
      sum[cl] = sqrt (sum[cl] / ((real) max_len));
    }
    minerr = sum[0];
    minerr_i = 0;
    //    message ("Target: %d %d", label, max_len);
    for (int cl=0; cl<n_classes; cl++) {
      //      message ("%d %f", cl, sum[cl]);
      if (sum[cl] < minerr) {
        minerr = sum[cl];
        minerr_i = cl;
      }
    }
    if (minerr_i!=label) {
      class_err ++;
    }
    if (error_buffer) {
      FreeSampleBuffer (error_buffer);
    }
  }
  
  stat_class_err =  100.0f*((real)class_err)/((real)test_data->n);
  message ("Classification Error: %f%%%", stat_class_err);
  printf ("%d %d %d %d ", (big_iter+1)*n_iter, n_hidden, n_hidden_b, rbf_layer);
  printf ("%d %d %d ", delay, look_ahead, embedding_dimension);
  printf ("%f %d %f %f %f", learning_rate, prior, stat_rms_train, stat_rms_test, stat_class_err);
  printf (" #TOPLOT\n"); //mark this line for grepping
  } /* bigiter */  

  /** Cleaning up **/
  for (int i=0; i<n_predictors; i++) {
    DeletePredictor (p[i]);
  }
  FreeM (p);
  FreeBufferList (train_data);
  FreeBufferList (test_data);

	message ("\nExiting\n");
	return 0;
}



