/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* VER: $Id: PredictorCluster.c,v 1.1 2003/08/21 02:08:01 olethros Exp $ */

#include "PredictorCluster.h"

PredictorCluster* NewPredictorCluster (int n_inputs, int d, int T, int look_ahead, int n_clusters)
{
	PredictorCluster* pc = NULL;
	if ((pc=AllocM (PredictorCluster, 1))==NULL) {
		return NULL;
	}
	pc->n_clusters = n_clusters;
	pc->predictor = AllocM (Predictor*, n_clusters);
	pc->e = AllocM (real, n_clusters);
	pc->d = d;
	pc->T = T;
	pc->look_ahead = look_ahead;
	for (int i=0; i<n_clusters; i++) {
		pc->predictor[i] = NewANNPredictor (n_inputs, d, T, look_ahead);
	}

	return pc;
}


int DeletePredictorCluster (PredictorCluster* pred_clus)
{
	for (int i=0; i<pred_clus->n_clusters; i++) {
		DeletePredictor (pred_clus->predictor[i]);
	}
	FreeM (pred_clus->predictor);
	FreeM (pred_clus->e);
	FreeM (pred_clus);
	return 0;
}	


int PredictorCluster_AddHiddenLayer (PredictorCluster* pred, int n_units)
{
	int e = 0;
	for (int i=0; i<pred->n_clusters; i++) {
		e+=Predictor_AddHiddenLayer (pred->predictor[i], n_units);
	}
	return e;
}

int PredictorCluster_AddRBFHiddenLayer (PredictorCluster* pred, int n_units)
{
	int e = 0;
	for (int i=0; i<pred->n_clusters; i++) {
		e+=Predictor_AddRBFHiddenLayer (pred->predictor[i], n_units);
	}
	return e;
}

void PredictorCluster_SetLearningRate (PredictorCluster* pred, real a)
{
	for (int i=0; i<pred->n_clusters; i++) {
		Predictor_SetLearningRate (pred->predictor[i], a);
	}
}

void PredictorCluster_SetBatchMode (PredictorCluster* pred, bool batch)
{
	for (int i=0; i<pred->n_clusters; i++) {
		Predictor_SetBatchMode (pred->predictor[i], batch);
	}
}

void PredictorCluster_BatchAdapt (PredictorCluster* pred)
{
	Serror ("No batch adaptation for predictor clusters\n");
}


void PredictorCluster_SetOutputsToTanH (PredictorCluster* pred)
{
	for (int i=0; i<pred->n_clusters; i++) {
		Predictor_SetOutputsToTanH (pred->predictor[i]);
	}
}

int PredictorCluster_Init (PredictorCluster* pred)
{
	int e = 0;
	for (int i=0; i<pred->n_clusters; i++) {
		e += Predictor_Init (pred->predictor[i]);
	}
	return e;
}

/* Functionality Interface */
real PredictorCluster_Input(PredictorCluster* predictor, real* x)
{
	for (int i=0; i<predictor->n_clusters; i++) {
		Predictor_Input (predictor->predictor[i], x);
	}
	return 0.0f;
}


/* Train only the predictor with the lowest error */
real PredictorCluster_Train(PredictorCluster* predictor, real* x)
{
	Predictor** p = predictor->predictor;
	return Predictor_Train (p[PredictorCluster_ArgMinErr (predictor, x)], x);
}

real Predictor_Test(PredictorCluster* predictor, real* x)
{
	real* e = predictor->e;
	return e[PredictorCluster_ArgMinErr (predictor, x)];
}

real* PredictorCluster_GetOutput(PredictorCluster* predictor)
{
	return Predictor_GetOutput (predictor->predictor[0]);
}

int PredictorCluster_ArgMinErr(PredictorCluster* predictor, real* x)
{
	real* e = predictor->e;
	int n = predictor->n_clusters;
	Predictor** p = predictor->predictor;

	for (int i=0; i<n; i++) {
		e[i] = Predictor_Test (p[i], x);
	}
	
	real min_e= e[0];
	int arg_min_e = 0;
	for (int i=0; i<n; i++) {
		if (min_e<e[i]) {
			min_e = e[i];
			arg_min_e = i;
		}
	}

	return arg_min_e;
}

