#include "ANN.h"
#include <time.h>


/* Test program for ANN. Results show that it works in some cases, but not all. Possibly there is a requirement for a bias term. */

int main (int argc, char** argv) {
	ANN* ann = NULL;
	int i;
	int n_inputs = 10;
	int n_outputs = 2;

	real* inputs;
	real* targets;
	real** data;
	int n_examples = 10;
	int max_iter = 1000;
	bool rbf = false;
	real learning_rate = 0.001f;
	real learning_rate_decay = 1.00f;
	FILE* outfile = fopen ("output_log", "w");
	if (outfile == NULL) {
		Serror ("Could not open file\n");
	}

	for (i=1; i<argc; i++) {
		if (!(strcmp (argv[i], "--rbf"))) {
			printf ("Using RBFs for the first hidden layer");
			rbf = true;
		}
	}

	srand48(time(NULL));
	


	/* Set up the data */
	inputs = AllocM (real, n_inputs);
	//	targets = AllocM (real, n_outputs);
		
	data = AllocM (real*, n_examples);
	for (i=0; i<n_examples; i++) {
		int j;
		data[i] =  AllocM (real, (n_inputs + n_outputs));
		for (j=0; j<(n_inputs); j++) {
			data[i][j] = drand48()*2.0f - 1.0f;
		}
		int xi = (int) (drand48()*((real) n_outputs));
		printf ("%d ", xi);
		for (j=0; j<(n_outputs); j++) {
			real x;
			if (xi==j) {
				x = +1.0f;
			} else {
				x = -1.0f;
			}
			data[i][j+n_inputs] = x;
		}
	}

	/* Create the Neural Network */
	ann = NewANN (n_inputs, n_outputs);
	if (rbf)
		ANN_AddRBFHiddenLayer (ann, 8);
	else
		ANN_AddHiddenLayer (ann, 8);
	//	ANN_AddHiddenLayer (ann, 8);
	ANN_Init (ann);
	ANN_SetLearningRate (ann, learning_rate);
	ANN_SetOutputsToTanH (ann);
	


	message ("Training for %d iterations", max_iter);
	for (i=0; i<max_iter; i++) {
		int k;
		real sum = 0.0f;
		
		for (k=0; k<n_examples; k++) {
			targets = &data[k][n_inputs];
			sum +=  ANN_Train (ann, data[k], targets);
		}
		
		fprintf (outfile, "%f %f\n", sum/((real) n_examples), learning_rate);
		fflush (outfile);
		learning_rate *= learning_rate_decay;
	}


	DeleteANN (ann);
	for (i=0; i<n_examples; i++) {
		FreeM (data[i]);
	}
	FreeM (data);
	FreeM (inputs);
	//	FreeM (targets);

	fclose (outfile);

	message ("Test complete\n");
	return 0;
}
