/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
const char* TESTPRED_VERSION = "VER: $Id: TestPredictor.c,v 1.7 2003/08/28 00:09:04 olethros Exp $";
const char* VERSION_INFO = "TestPredictor 0.1\n\
Copyright (C) 2003 Christos Dimitrakakis\n\
TestPredictor comes with no WARRANTY,\n\
to the extent permitted by law.\n\
You may redistribute copies of LinearPredictor\n\
under the terms of the GNU General Public License.\n\
For more information about these matters\n\
see the files named COPYTING.\n";

const char* help = "\n\
  TestPredictor <options>\n\n\
  --inputs        \t Dimensionality of input\n\
  --delay         \t Delay of Data [1]\n\
  --dimension     \t Embedding dimension [32]\n\
  --ahead         \t Look-ahead [0]\n\
  --epochs        \t Number of epochs to train for [100]\n\
  --nhu           \t Number of hidden units [25]\n\
  --nhu2          \t Number of hidden units in the 2nd hidden layer [0]\n\
  --lr            \t Set learning rate [0.1]\n\
  --decay         \t Set learning rate decay [1.0]\n\
  --train         \t Training file\n\
  --average       \t Average for prediction\n\
  --help       -h \t Show this help\n\n\
  --version    -v \t Show version information and exit\n\n\
Report bugs to Christos Dimitrakakis <dimitrak@idiap.ch>\n";




#include "Predictor.h"
#include "SampleBuffer.h"
#include "psr.h"
#include <time.h>
#include <getopt.h>
#include <assert.h>
//#include <list.h>

#undef USE_HCNN

#define ControlParameter(i, maxi) ((1.6f * ((real) i)/((real) maxi) - 0.8f))
void Generate (Predictor* p, int step_size, char* fname, SAMPLE_BUFFER* sample);void Predict (Predictor* p, int step_size, char* fname, SAMPLE_BUFFER* sample);

int main (int argc, char** argv)
{

	Predictor* p;
	real* data = NULL;
  int l = 100; // for data generation
	int i;
	int n_iter = 25;//100;
	int iter;
	int window_size;
  int n_inputs = 1;
  int n_hidden = 25;
  int n_hidden_b = 0;
  bool rbf_layer = false;
  bool tanh_outputs = false;
  bool average = false;
  int delay = 1;
  int look_ahead = 0;
  int embedding_dimension = 32;
  real learning_rate = 0.01f;
  real learning_rate_decay = 1.0;//0.999f;
	FILE* outfile;
  bool read_sample = true;
  char* sample_name = "/home/olethros/data/sound/kalos.raw";
  char* test_sample_name = NULL;
  char* test2_sample_name = NULL;
  SAMPLE_BUFFER* sample = NULL;
  SAMPLE_BUFFER* test_sample = NULL;
  SAMPLE_BUFFER* test2_sample = NULL;
  int test_l;
  int test2_l;
  char* pred_fname = "pred.raw";



const char* help = "\n\
  TestPredictor <options>\n\n\
  --inputs        \t Dimensionality of input\n\
  --delay      -T \t Delay of Data [1]\n\
  --dimension  -d \t Embedding dimension [32]\n\
  --ahead      -a \t Look-ahead [0]\n\
  --epochs     -N \t Number of epochs to train for [100]\n\
  --rbf           \t Use RBFs for the first hidden layer\n\
  --tanh          \t Use Hyperbolic Tan at the output\n\
  --nhu           \t Number of hidden units [25]\n\
  --nhu2          \t Number of hidden units in the 2nd hidden layer [0]\n\
  --lr            \t Set learning rate [0.1]\n\
  --decay         \t Set learning rate decay [1.0]\n\
  --average       \t Average over outputs for prediction\n\
  --train         \t Training file\n\
  --test          \t Testing file\n\
  --help       -h \t Show this help\n\n\
  --version    -v \t Show version information and exit\n\n\
Report bugs to Christos Dimitrakakis <dimitrak@idiap.ch>\n";
  while (1) {
    int option_index = 0;
    static struct option long_options[] = {
      {"inputs", required_argument, 0, 0 }, //0
      {"delay", required_argument, 0, 'T'}, //1
      {"dimension", required_argument, 0, 'd'}, //2
      {"ahead", required_argument, 0, 'a'}, //3
      {"epochs", required_argument, 0, 'N'}, //4
      {"nhu", required_argument, 0, 0}, //5
      {"nhu2", required_argument, 0, 0}, //6
      {"lr", required_argument, 0, 0}, //7
      {"decay", required_argument, 0, 0}, //8
      {"train", required_argument, 0, 0}, //9
      {"rbf", no_argument, 0, 0}, //10
      {"tanh", no_argument, 0, 0}, //11
      {"average", no_argument, 0, 0}, //12
      {"version", no_argument, 0, 'v'}, //13
      {"help", no_argument, 0, 'h'}, //14
      {"test", required_argument, 0, 0}, //15
      {"test2", required_argument, 0, 0}, //16
      {0, 0, 0, 0}
    };

    int c = getopt_long (argc, argv, "i:d:a:N:T:vh",
                         long_options, &option_index);
    if (c == -1) {
      break;
    }
    switch (c) {
    case 0: /* This is the case for long arguments */
#ifdef DEBUG_ARGUMENTS
      printf ("option %s", long_options[option_index].name);
      if (optarg)
        printf (" with arg %s", optarg);
      printf ("\n");
#endif

      switch(option_index) {

      case 5: n_hidden = atoi (optarg); break;
      case 6: n_hidden_b = atoi (optarg); break;
      case 7: learning_rate = atof (optarg); break;
      case 9: sample_name = optarg; break;
      case 10: printf ("Using RBFs\n"); rbf_layer = true; break;
      case 11:
        printf ("Setting activation to tanh\n");
        tanh_outputs = true;
        break;
      case 12:
        printf ("Averaging for prediction\n");
        average = true;
        //        break;
      case 13: printf ("Version\n"); break;
      case 14: printf ("HELP\n"); break;
      case 15: test_sample_name = optarg; break;
      case 16: test2_sample_name = optarg; break;
      default :
        Swarning ("This should never happen\n");
      }
      break;
      
    case 'i': n_inputs = atoi (optarg); break;
    case 'd': embedding_dimension = atoi (optarg); break;
    case 'a': look_ahead = atoi (optarg); break;
    case 'N': n_iter = atoi (optarg); break;
    case 'T': delay = atoi (optarg); break;

    case 'v':
      printf ("%s", TESTPRED_VERSION);
      break;
    case 'h':
      printf("%s",help);
      exit(0);
      break;
    case '?':
      printf ("Ambiguous match\n");
      break;
    case ':':
      printf ("Missing parameter\n");
      break;
    default:
      printf ("?? getopt returned character code 0%o ??\n", c);
    }
  }

  printf ("%s\n", TESTPRED_VERSION);

  window_size = embedding_dimension * delay;
  printf ("Window size : %d\n", window_size);
  printf ("Embedding Dimension : %d\n", embedding_dimension);
  printf ("Delay : %d\n", delay);

  srand48(time(NULL));

  if (read_sample) {
    sample = LoadSampleBuffer (sample_name);
    if (sample == NULL) {
      Serror ("Failed to read %s\n", sample_name);
      return -1;
    }
    l = sample->length;
    sample = ExtendBuffer (sample, window_size + delay + look_ahead);
    
    data = sample->x;
    if (test_sample_name) {
      test_sample = LoadSampleBuffer (test_sample_name);
      test_l = test_sample->length;
      test_sample = ExtendBuffer (test_sample, window_size + delay + look_ahead);
    } else {
      test_sample = NULL;
    }
    if (test2_sample_name) {
      test2_sample = LoadSampleBuffer (test2_sample_name);
      test2_l = test2_sample->length;
      test2_sample = ExtendBuffer (test2_sample, window_size + delay + look_ahead);
    } else {
      test_sample = NULL;
    }

  } else {
    printf ("Error !");
  }
  

#ifdef USE_HCNN
	p = NewHCNNPredictor (n_inputs, embedding_dimension, delay, look_ahead);
#else
	p = NewANNPredictor (1, embedding_dimension, delay, look_ahead);
#endif
  if (n_hidden > 0) {
    if (rbf_layer) {
      Predictor_AddRBFHiddenLayer (p, n_hidden);
    } else {
      Predictor_AddHiddenLayer (p, n_hidden);
    }
    if (n_hidden_b > 0) {
      Predictor_AddHiddenLayer (p, n_hidden_b);
    }
  }

	Predictor_Init (p);
	Predictor_SetLearningRate (p, learning_rate);
  if (tanh_outputs) {
    Predictor_SetOutputsToTanH (p);
  }

  /* Training loop */
  printf ("Training for %d iterations\n", n_iter);
  outfile = fopen ("pred_rms", "w");
  if (outfile) {
    for (iter=0; iter<n_iter; iter++) {
      real sum = 0.0f;
      int max_len = l;
      /* learn by one, or by delay */
      for (i=0; i<max_len; i+=delay) {
#ifdef USE_HCNN
        real c = ControlParameter (i, max_len);
        sum += Predictor_HCNNTrain (p, &data[i], c);
#else
        sum += Predictor_Train (p, &data[i]);
#endif
      }
      sum = sqrt (sum / ((real) max_len));
      printf ("rms: %f\n", sqrt(sum));
      fprintf (outfile, "%f\n", sum);
      
      learning_rate *= learning_rate_decay;
      Predictor_SetLearningRate (p, learning_rate);
    }
    
    fclose (outfile);
  }


  printf ("\nTesting on train data\n");
  Predict (p, 1, pred_fname, sample);
  if (test_sample) {
    printf ("\nTesting on test data 1\n");
    Predict (p, 1, pred_fname, test_sample);
  }
  if (test2_sample) {
    printf ("\nTesting on test data 2\n");
    Predict (p, 1, pred_fname, test2_sample);
  }

	printf ("\nGenerating\n\n");
  // Generate (p, 1, "/home/olethros/data/sound/gen1.raw", sample);
  Generate (p, delay, "gen2.raw", sample);

	DeletePredictor (p);
  if (read_sample) {
    FreeSampleBuffer (sample);
  } else {
    FreeM (data);
  }

	printf ("\nExiting\n");
	return 0;
}


void Generate (Predictor* p, int step_size, char* fname, SAMPLE_BUFFER* sample) {
  int i;
  SAMPLE_BUFFER* gen_sample_buffer;


  real* data = sample->x;
  FILE* gen_file = NULL;
  int window_size = p->d * p->T;

  gen_sample_buffer = NewSampleBuffer (sample->length + window_size + p->look_ahead + p->T ,0);
  real* buffer = gen_sample_buffer->x;

  for (i=0; i<(gen_sample_buffer->length); i++) {
    //   if (i<window_size) {
    buffer[i] = -1.0;//data[i];
      //} else {
      // buffer[i] = 0.0;
      // }
  }
  
  buffer[0] = data[0];

  /* Feedback the buffer input */
  int max_len = sample->length;//x - window_size - p->look_ahead - p->T;
  for (i=0; i<max_len; i+=step_size) {
    real* y;
    int j;
#ifdef USE_HCNN
    real c = ControlParameter (i, max_len);
    Predictor_HCNNInput (p, &buffer[i], c);
#else
    Predictor_Input (p, &buffer[i]);
#endif
    y = Predictor_GetOutput (p);
    for (j=0; j<p->T; j++) {
      buffer[i+window_size+j-p->T+1] = y[j];// * pow(0.9, (float) j);
    }
  }
  gen_file = fopen (fname, "w");
  if (gen_file) {
    ScaleSampleBuffer (gen_sample_buffer);
    SaveSampleBuffer (gen_sample_buffer, gen_file);
    fclose (gen_file);
  } else {
    Swarning ("Did not save generated data in %s\n", fname);
  }
  FreeSampleBuffer (gen_sample_buffer);

}

void Predict (Predictor* p, int step_size, char* fname, SAMPLE_BUFFER* sample)
{
  real* data;
  FILE* pred_file = NULL;

  int i;
  int max_len = sample->length - p->T - p->look_ahead;
  int delay = p->T;
  int l = max_len;
  int look_ahead = p->look_ahead;
  int window_size = delay*p->d;
  real sumt = 0.0f;
  data = sample->x;

  real* buffer;
  SAMPLE_BUFFER* out_sample_buffer;
  /* Create a prediction buffer */
  out_sample_buffer = NewSampleBuffer (l  + window_size + delay + look_ahead, CLEAR_SAMPLE_BUFFER);
  buffer = out_sample_buffer->x;


	printf ("Testing\n");
  for (i=0; i<max_len; i+=step_size) {
    real* y;
    real e;
    int j;
#ifdef USE_HCNN
    real c = ControlParameter (i, max_len);
    Predictor_HCNNInput (p, &data[i], c);
#else
    Predictor_Input (p, &data[i]);
#endif
    y = Predictor_GetOutput (p);
    for (j=0; j<delay; j++) {
      buffer[i+j] += y[j];
      e = data[window_size+i-p->T+1+j] - y[j];
      e = e*e;
      sumt += e; 
    }
  }
  printf ("mse: %f\n", sqrt(sumt / ((real) max_len)));
  
  pred_file = fopen (fname, "w");
  if (pred_file) {
    ScaleSampleBuffer (out_sample_buffer);
    SaveSampleBuffer (out_sample_buffer, pred_file);
    fclose (pred_file);
  }
  
    
  FreeSampleBuffer (out_sample_buffer);
}
