/* -*- Mode: C++;  -*- */
/* "VER: $Id: Windowing.c,v 1.3 2003/09/06 17:26:18 olethros Exp olethros $ */

#include "Windowing.h"
#include "debug.h"
#include "psr.h"

const SAMPLE_BUFFER* RectangularWindow::Cut (const SAMPLE_BUFFER* input, TimeStamp start_time, TimeStamp attack)
{
	TimeStamp l = N;
	if (attack<=0) {
		Swarning ("Illegal arguments for Cut()\n");
		return NULL;
	}
     
	CopyBufferRange (tmp, input, start_time, l);
	VolumeRamp (tmp, 0, attack, 0.0, 1.0);
	VolumeRamp (tmp, N - attack - 1, N - 1, 1.0 , 0.0);

	return tmp;
}


SAMPLE_BUFFER* RectangularWindow::Cut (const SAMPLE_BUFFER* input, TimeStamp start_time, TimeStamp end_time, TimeStamp attack)
{
	TimeStamp l = end_time - start_time;
	if ((start_time>=end_time)||(attack<=0)) {
		Swarning ("Illegal arguments for Cut()\n");
		return NULL;
	}

	SAMPLE_BUFFER* sb = CopyBufferRange (input, start_time, l);
	VolumeRamp (sb, 0, attack, 0.0, 1.0);
	VolumeRamp (sb, (l - attack - 1), (l -1), 1.0 , 0.0);

	return sb;
}


const SAMPLE_BUFFER* HammingWindow::Cut (const SAMPLE_BUFFER* input, TimeStamp start_time, TimeStamp attack)
{
	if (attack<=0) {
		Swarning ("Illegal arguments for Cut()\n");
		return NULL;
	}

	Filter (input->x, tmp->x, N);

	return tmp;
}




SAMPLE_BUFFER* HammingWindow::Cut (const SAMPLE_BUFFER* input, TimeStamp start_time, TimeStamp end_time,  TimeStamp attack)
{
	TimeStamp l = end_time - start_time;
	if ((l<=0)||(attack<=0)) {
		Swarning ("Illegal arguments for Cut()\n");
		return NULL;
	}

	SAMPLE_BUFFER* sb = CopyBufferRange (input, start_time, l);
	Filter (input->x, sb->x, l);

	return sb;
}

void HammingWindow::Filter (float* src, float* dst, TimeStamp l)
{
	register float theta = 2.0 * M_PI / ((float) (l-1));
	register float u = 0.0f;
	register float M_CPI = 2.0 * M_PI;

	if (l<=1) {
		Serror ("A bit too short, this window\n");
		return;
	}

	for (TimeStamp i=0; i<l; i+=1) {
		*dst++ = *src++ * (0.54 - 0.46 * cos (u));
		u += theta;
		if (u > M_CPI) {
			u -= M_CPI;
		}
	}
}
