#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include "SampleBuffer.h"
#include "psr.h"
#include "debug.h"
#include "real.h"

#define SHOW_STATISTICS

real test_alignment (int n_seq, long length, real lr);

int main (void)
{
  int i;
  int j;
  real lr;
  srand48(time(NULL));
  
  for (i=2; i<=256; i=i*2) {
    lr = 0.001f;
    for (j=0; j<10; j++) {
      real v = test_alignment (i, 10000, lr);
#ifdef SHOW_STATISTICS
      printf ("%d %f %f\n", i, lr, v);
      //printf ("%f ",v);
#endif
      lr = lr * 2.0f;
    }
    printf ("\n");
  }
  
  return 0;

}

#ifdef SHOW_ALIGNMENT
#ifdef SHOW_STATISTICS
#error "Cannot output both alignment and statistics at the same time"
#endif
#endif

#ifndef SHOW_ALIGNMENT
#ifndef SHOW_STATISTICS
#error "Define one of the data-display options!"
#endif
#endif

real test_alignment (int n_seq, long length, real lr)
{
  SAMPLE_BUFFER** seq = NULL;
  real* x = NULL;
  real* ix = NULL;
  real* dx = NULL;
  real* d = NULL;
  real variance = 0.0f;
  real mean = 0.0f;
  int t = 0;
  int i = 0;
  int j = 0;


  seq = AllocM (SAMPLE_BUFFER*, n_seq);
  d = AllocM (real, n_seq);
  x = AllocM (real, n_seq);
  ix = AllocM (real, n_seq);
  dx = AllocM (real, n_seq);

  seq[0] = CreatePseudoRandomSequence (length, 0.1f);
  //  seq[0] = CreateSine (length, 44100.0f, 1000.0f, 1.0f);
  //  seq[0] = NewSampleBuffer (length, CLEAR_SAMPLE_BUFFER);
  //  for (t=0; t<length; t+=50) {
  //    seq[0]->x[t] = 1.0f;
  //  }
  for (i=1; i<n_seq; i++) {
    long delay = (int) (drand48()*32.0f);
    //    fprintf (stderr, "Delay [%d] = %ld\n", i, delay);
    d[i] = 0.0f;
    x[i] = 0.0f;
    seq[i] = NewSampleBuffer (length, CLEAR_SAMPLE_BUFFER);
    for (j=0; j < (seq[i]->length - delay); j++) {
      seq[i]->x[j] = seq[0]->x[j+delay];
    }
  }

  for (i=0; i<n_seq; i++) {
    AddUniformNoise (seq[i], 0.01f);
  }

  for (t=0; t<length; t++) {
    real sum = 0.0f;
    for (i=0; i<n_seq; i++) {
      real y;
      int pos = t+ (int) d[i];
      if ((pos<0) || (pos>=length)) {
	y = 0.0f;
      } else {
	y = seq[i]->x[pos];
      }
      dx[i] = 0.1*(ix[i]-x[i]);
      ix[i] = ix[i]*0.9 + y;
      x[i] = x[i]*0.9 + 0.1*ix[i];


      sum += y;
    }

    for (i=0; i<n_seq; i++) {
      for (j=0; j<n_seq; j++) {
	real mul = x[i]*x[j];
	real diff = dx[i]-dx[j];
	d[i] += lr*diff*mul;
      }
    }

#ifdef SHOW_STATISTICS
    if ((length-t)>1000) {
      real smean = 0.0f;
      real svar = 0.0f;
      for (i=0; i<n_seq; i++) {
	smean += x[i];
      }
      smean = smean/((real) n_seq);
      for (i=0; i<n_seq; i++) {
	svar = (x[i]-smean);
      }
      mean += smean;
      variance += (svar*svar);
      
    }
	  
    
#endif

#ifdef SHOW_ALIGNMENT
    for (i=0; i<n_seq; i++) {
      printf ("%f ",x[i]);
      printf ("%f ",d[i]);
    }
    printf ("%f\n",sum/n_seq);
#endif

    


  } /* for t<length */

  for (i=0; i<n_seq; i++) {
    FreeSampleBuffer (seq[i]);
    seq[i] = NULL;
  }
  FreeM (seq);
  FreeM (d);
  FreeM (x);
  FreeM (dx);
  variance = 0.001f*variance;
  fprintf(stderr, "%d %f %f\n", n_seq, lr, variance);
  return variance;
}
