/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* "VER: $Id: modulator.h,v 1.3 2003/08/28 03:37:28 olethros Exp $" */

#ifndef MODULATOR_H
#define MODULATOR_H

#include <math.h>
#include <assert.h>
#include "SampleBuffer.h"

/* Modulators are implemented via function pointers.  The four floats
   are the four controllable parameters of the modulator.  I have not
   settled yet on a formal design for how modulation should be done in
   general, but the general principle is that all effects that can be
   called with a Modulator as argument should call
   modulator->fun(modulator, x) every N processing steps. N is left to
   the discretion of the effect's designer, I guess, though it should
   be 1 in most cases. This does not make for the fastest design, but
   it is the most flexible.

   The other alternative is to be able to use samples as a modulator,
   but in our case we simply define that possibility through the
   mod_sample modulator.

   Anyway, all modulators share the four a,b,c,d float parameters and
   some are more versatile.
*/

typedef struct Modulator_ {
  float (*fun) (struct Modulator_* mod, float x);
  float a, b, c, d; //the 4 standard parameters for fast access
} Modulator;

/* Child structures of Modulator - this one can be useful but one should be careful in its use. */
typedef struct VirtualModulator_ {
  struct Modulator_ modulator;
  void* u;
} VirtualModulator;

/* Use this one to have an effect manipulated by a sample */
typedef struct SampleModulator_ {
  struct Modulator_ modulator;
  SAMPLE_BUFFER* u;
} SampleModulator;

/* This one is to be used to set up a chain of modulators. Use with care.*/
typedef struct ModModulator_ {
  struct Modulator_ modulator;
  struct Modulator_* u;
} ModModulator;

float mod_sine (Modulator* mod, float x);
float mod_exp_decay (Modulator* mod, float l);
float mod_sample (VirtualModulator* mod, float x);

#define MODFN(x,y) (x->fun(x,((float)y)))

#endif /* MODULATOR_H */
