/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* $Id: seq2multi.c,v 1.1 2002/10/31 18:25:36 olethros Exp $ */

/*
  seq2multi - By Christos Dimitrakakis, 2002 - Send complaints to dimitrak@idiap.ch

  DESCRIPTION
  
  - This simple program is used to convert a single multi-sequence
  file into many sub-sequence, single-class files, that can be used to
  read into a MatDataSet with the following method:

    MatDataSet(char **input_filenames, char **target_filenames, int n_files_,
               int max_load=-1, bool binary_mode=false);

  - It convers a file of the type
      ....
      <INPUT DATA> <CLASS> <SEQ>
      ....
  
  To a set of files:
    prefix.inputs.<class>_<seq>
    prefix.targets.<seq>

  A list of files is also saved in 
    prefix.list

  - The format of prefix.inputs is
      N_ROWS N_COLUMNS
      ....
      <INPUT_DATA>
      ....

  - The format of prefix.targets is
      N_ROWS N_COLUMNS
      ....
      <TARGETS>
      ....

  - The format of prefix.list is
      N_ROWS N_COLUMNS
      ....
      <INPUT_FILE> <TARGET_FILE>
      ....

  SEE ALSO
  - tag2num.c
  - datafile.c
  - LinkedList.c

*/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include "datafile.h"
#include "LinkedList.h"



int main(int argc, char* argv[])
{
  char* InFileName = NULL;
  char* OutPrefix = NULL;
  DATAFILE* DataFile;
  int column;
  int seqcolumn;

  if (argc!=5) {
	fprintf(stderr,"USAGE: seq2multi <infile> <prefix> <class column> <sequence column>\n");
	return -1;
  }


  InFileName = argv[1];
  OutPrefix = argv[2];
  column = atoi(argv[3]);
  seqcolumn = atoi(argv[4]);
  
  
  printf("Replacing column %d of file %s\n",column,InFileName);
  if ((DataFile = ScanDataFile(InFileName))==NULL) {
    Serror("Could Not create datafile placeholder\n");
  }
  
  /* Actually, all the functionality is implemented in datafile.c!
     MWUAHAHAHAHA - hope you have the source code for that! */
  DisruptSequenceData(DataFile, OutPrefix, column, column+1);

  CloseFile(DataFile->InputFile);

  FreeM(DataFile);

  return 0;

}





