/* -*- Mode: C++; -*- */
const char* CLUSPRED_VERSION = "VER: $Id: sequence_classifier.c,v 1.5 2003/09/05 01:43:47 olethros Exp $";
const char* VERSION_INFO = "Sequence Classifier 0.1\n\
Copyright (C) 2003 Christos Dimitrakakis\n\
Sequence Classifier comes with no WARRANTY,\n\
to the extent permitted by law.\n\
You may redistribute copies of PredictorClassifier\n\
under the terms of the GNU General Public License.\n\
For more information about these matters\n\
see the files named COPYTING.\n";

const char* help = "\n\
  sequence_classifier <options>\n\n\
  --inputs        \t Dimensionality of input\n\
  --delay         \t Delay of Data [1]\n\
  --dimension     \t Embedding dimension [32]\n\
  --ahead         \t Look-ahead [0]\n\
  --epochs        \t Number of epochs to train for [100]\n\
  --nhu           \t Number of hidden units [25]\n\
  --nhu2          \t Number of hidden units in the 2nd hidden layer [0]\n\
  --lr            \t Set learning rate [0.1]\n\
  --decay         \t Set learning rate decay [1.0]\n\
  --train         \t Training file\n\
  --labels        \t A file containing the list of labels\n\
  --average       \t Average for prediction\n\
  --states        \t Display cluster states\n\
  --batch         \t Use batch mode for training\n\
  --help       -h \t Show this help\n\n\
  --version    -v \t Show version information and exit\n\n\
Report bugs to Christos Dimitrakakis <dimitrak@idiap.ch>\n";


#include <time.h>
#include <getopt.h>
#include <assert.h>
#include "PredictorCluster.h"
#include "SampleBuffer.h"
#include "SegmentedSample.h"
#include "datafile.h"
#include "psr.h"

real test_model (PredictorCluster** p, IOExampleList* test_data, int window_size, int delay, int cutoff, int n_classes, int prior);


int main (int argc, char** argv)
{
	int n_iter = 25;//100;
	int window_size;
	int cutoff;
	int n_clusters = 1;
	int n_classes = 0;
	int n_inputs = 1;
	int n_hidden = 25;
	int n_hidden_b = 0;
	bool rbf_layer = false;
	bool tanh_outputs = false;
	bool average = false;
	int delay = 1;
	int look_ahead = 0;
	int embedding_dimension = 32;
	real learning_rate = 0.01f;
	real learning_rate_decay = 1.0;//0.999f;
	int prior = 0;
	bool batch_mode = false;
	bool states = false;
	char* train_fname = NULL;
	char* test_fname = NULL;
	char* label_fname = NULL;
	real stat_rms_train = 0.0f;
	real stat_rms_test = 0.0f;
	real stat_class_err_train = 0.0f;
	real stat_class_err_test = 0.0f;

	/* Labels */
	LINKEDLIST* labels;

	/* File pairs (inputs and targets) */
	FilePair train_pairs;
	FilePair test_pairs;

	IOExampleList* train_data;
	IOExampleList* test_data;

	PredictorCluster** p;  
	int n_predictors = 0;




	const char* help = "\n\
  sequence_classifier <options>\n\n\
  --n_clusters    \t Number of clusters per class [1]\n\
  --prior         \t Also use a general predictor\n\
  --inputs        \t Dimensionality of input\n\
  --labels        \t List of labels\n\
  --classes       \t Number of Classes\n\
  --delay      -T \t Delay of Data [1]\n\
  --dimension  -d \t Embedding dimension [32]\n\
  --ahead      -a \t Look-ahead [0]\n\
  --epochs     -N \t Number of epochs to train for [100]\n\
  --rbf           \t Use RBFs for the first hidden layer\n\
  --tanh          \t Use Hyperbolic Tan at the output\n\
  --nhu           \t Number of hidden units [25]\n\
  --nhu2          \t Number of hidden units in the 2nd hidden layer [0]\n\
  --lr            \t Set learning rate [0.1]\n\
  --decay         \t Set learning rate decay [1.0]\n\
  --average       \t Average over outputs for prediction\n\
  --train         \t Training file\n\
  --test          \t Testing file\n\
  --help       -h \t Show this help\n\n\
  --version    -v \t Show version information and exit\n\n\
Report bugs to Christos Dimitrakakis <dimitrak@idiap.ch>\n";
	while (1) {
		int option_index = 0;
		static struct option long_options[] = {
			{"inputs", required_argument, 0, 0 }, //0
			{"delay", required_argument, 0, 'T'}, //1
			{"dimension", required_argument, 0, 'd'}, //2
			{"ahead", required_argument, 0, 'a'}, //3
			{"epochs", required_argument, 0, 'N'}, //4
			{"nhu", required_argument, 0, 0}, //5
			{"nhu2", required_argument, 0, 0}, //6
			{"lr", required_argument, 0, 0}, //7
			{"decay", required_argument, 0, 0}, //8
			{"train", required_argument, 0, 0}, //9
			{"rbf", no_argument, 0, 0}, //10
			{"tanh", no_argument, 0, 0}, //11
			{"average", no_argument, 0, 0}, //12
			{"version", no_argument, 0, 'v'}, //13
			{"help", no_argument, 0, 'h'}, //14
			{"test", required_argument, 0, 0}, //15
			{"classes", required_argument, 0, 0}, //16
			{"prior", no_argument, 0, 0},//17
			{"batch", no_argument, 0, 0},//18
			{"n_clusters", required_argument, 0, 0}, //19
			{"states", no_argument, 0, 0}, //20
			{"labels", required_argument, 0, 0}, //21
			{0, 0, 0, 0}
		};

		int c = getopt_long (argc, argv, "i:d:a:N:T:vh",
							 long_options, &option_index);
		if (c == -1) {
			break;
		}
		switch (c) {
		case 0: /* This is the case for long arguments */
#ifdef DEBUG_ARGUMENTS
			printf ("option %s", long_options[option_index].name);
			if (optarg)
				printf (" with arg %s", optarg);
			printf ("\n");
#endif

			switch(option_index) {

			case 5: n_hidden = atoi (optarg); break;
			case 6: n_hidden_b = atoi (optarg); break;
			case 7: learning_rate = atof (optarg); break;
			case 8: learning_rate_decay = atof (optarg); break;
			case 9: train_fname = optarg; break;
			case 10: message ("Using RBFs\n"); rbf_layer = true; break;
			case 11:
				message ("Setting activation to tanh\n");
				tanh_outputs = true;
				break;
			case 12:
				message ("Averaging for prediction\n");
				average = true;
				//        break;
			case 13: printf ("Version\n"); break;
			case 14: printf ("HELP\n"); break;
			case 15: test_fname = optarg; break;
			case 16: n_classes = atoi (optarg); break;
			case 17: prior = 1; break;
			case 18: batch_mode = true; break;
			case 19: n_clusters = atoi (optarg); break;
			case 20: states = true; break;
			case 21: label_fname = optarg; break;
			default :
				Swarning ("This should never happen\n");
			}
			break;
      
		case 'i': n_inputs = atoi (optarg); break;
		case 'd': embedding_dimension = atoi (optarg); break;
		case 'a': look_ahead = atoi (optarg); break;
		case 'N': n_iter = atoi (optarg); break;
		case 'T': delay = atoi (optarg); break;

		case 'v':
			printf ("%s", CLUSPRED_VERSION);
			break;
		case 'h':
			printf("%s",help);
			exit(0);
			break;
		case '?':
			Serror ("Ambiguous match\n");
			exit (-1);
			break;
		case ':':
			Serror ("Missing parameter\n");
			exit (-1);
			break;
		default:
			printf ("?? getopt returned character code 0%o ??\n", c);
		}
	}

	labels = LoadHTKLabels (label_fname);

	message ("%s", CLUSPRED_VERSION);
	message ("Labels: %s", label_fname);
	message ("N_ITER:%d",  n_iter);
	message ("n_clusters:%d", n_clusters);
	message ("n_classes:%d", n_classes);
	message ("n_inputs:%d", n_inputs);
	message ("n_hidden:%d",  n_hidden);
	message ("n_hiddenb:%d",  n_hidden_b);
	message ("rbf_layer %d", rbf_layer);
	message ("tanh %d", tanh_outputs);
	message ("average %d",  average);
	message ("T %d", delay);
	message ("l %d", look_ahead);
	message ("d %d", embedding_dimension);
	message ("lr %f", learning_rate);
	message ("lrd %f", learning_rate_decay);
	message ("prior %d", prior);

	srand48 (time (NULL));

	window_size = embedding_dimension * delay ;
	cutoff = window_size + look_ahead + delay;


	if (n_classes != Length (labels)) {
		Swarning ("Number of classes (%d) differs from number of labels(%d), setting them to be equal\n", n_classes, Length (labels));
		n_classes = Length (labels);
	}
  
	n_predictors = n_classes + prior;  
  
	/*** Initialise predictors ***/
	p = AllocM (PredictorCluster*, n_predictors);
	for (int i=0; i<n_predictors; i++) {
		if ((prior) && (i==n_classes)) {
			p[i]=  NewPredictorCluster (n_inputs, embedding_dimension, delay, look_ahead, 1);
		} else {
			p[i]=  NewPredictorCluster (n_inputs, embedding_dimension, delay, look_ahead, n_clusters);
		}
		if (n_hidden > 0) {
			if (rbf_layer) {
				PredictorCluster_AddRBFHiddenLayer (p[i], n_hidden);
			} else {
				PredictorCluster_AddHiddenLayer (p[i], n_hidden);
			}
			if (n_hidden_b > 0) {
				PredictorCluster_AddHiddenLayer (p[i], n_hidden_b);
			}
		}
		PredictorCluster_Init (p[i]);
		PredictorCluster_SetLearningRate (p[i], learning_rate);
		PredictorCluster_SetBatchMode (p[i], batch_mode);
		if (tanh_outputs) {
			PredictorCluster_SetOutputsToTanH (p[i]);
		}
	}


	/*** Load Data ***/

	/* OK - check out the lists */
	if (readFileList(train_fname, &train_pairs)) {
		Serror("Could not read training list\n");
		exit (-1);
	}
	if (readFileList(test_fname, &test_pairs)) {
		Serror("Could not read testing list\n");
		exit (-1);
	}

	int extra_room;
	if (prior) {
		extra_room = cutoff*2;
	} else {
		extra_room = cutoff;
	}
	train_data = LabelSegments (&train_pairs, labels, extra_room);
	test_data = LabelSegments (&test_pairs, labels, extra_room);

	SelectFirstExample (train_data);
	{
		float rate = (float) train_data->input->rate;
		message ("Sampling rate: %fms, %fHz", 1000.0/rate, rate);
		message ("Delay: %fms, %fHz", 1000.0*((float)delay)/rate, rate/((float)delay));
		message ("d: %fms, %fHz", 1000.0*((float)delay*embedding_dimension)/rate, rate/((float)delay*embedding_dimension));
		message ("a: %fms, %fHz", 1000.0*((float)delay+look_ahead)/rate, rate/((float)delay+look_ahead));
	}

    


	/*** Training Loop ***/
	for (int big_iter=0; big_iter<20; big_iter++) {
		message ("Training for %d iterations", n_iter);
		if (prior) {
			message ("(First training the prior PredictorCluster)");

			for (int iter=0; iter<n_iter; iter++) {
				real sum = 0.0f;
				int total_length = 0;

				/* learn by one, or by delay */
				SelectFirstExample (train_data);
				while (train_data->current) {
					real* data = train_data->input->x;
					int max_len = train_data->input->length - cutoff;
					for (int i=0; i<max_len; i+=delay) {
						sum += PredictorCluster_Train (p[n_classes], &data[i]);
					}
					total_length += max_len;
					SelectNextExample (train_data);
				}
				sum = sqrt (sum / ((real) total_length));
				if (iter == n_iter-1) {
					message ("rms: %f", sum);
					stat_rms_train = sum;
				}
				if (batch_mode) {
					PredictorCluster_BatchAdapt (p[n_classes]);
				}
			}
		}
  


		SAMPLE_BUFFER* error_buffer = NULL;

		if (prior) {
			error_buffer = NewSampleBuffer ();
		}

		message ("(Training classes)");

		for (int iter=0; iter<n_iter; iter++) {
			real sum = 0.0f;
			int total_length = 0;
    
			/* learn by one, or by delay */
			SelectFirstExample (train_data);
			while (train_data->current) {
				real* data = train_data->input->x;
				int max_len = train_data->input->length - cutoff;
				int err_len = max_len + delay;
				int label = train_data->target;

				assert (( (label>=0) && (label<n_classes) ));

				if (prior) {
					RequestBufferLength (error_buffer, err_len);
					for (int i=0; i<max_len; i+=delay) {
						PredictorCluster_Input (p[n_classes], &data[i]);
						real* y = PredictorCluster_GetOutput (p[n_classes]);
						real* b = error_buffer->x;
						for (int t=0; t<delay; t++) {
							int index = window_size+i-p[n_classes]->T+t+1;
							if ((i+t)>=error_buffer->length) {
								Serror ("Buffer Length %ld, trying to write to %d+%d=%d (c:%d)\n", error_buffer->length, i, t, i+t, cutoff);
							}
							b[i+t] = data[index] - y[t];
						}
					}
					max_len = max_len - cutoff;
					data = error_buffer->x;
				}
				for (int i=0; i<max_len; i+=delay) {
					real e = PredictorCluster_Train (p[label], &data[i]);
					sum += e;
				}
				total_length += max_len;
				SelectNextExample (train_data);
			}

			if (error_buffer) {
				FreeSampleBuffer (error_buffer);
			}

			sum = sqrt (sum / ((real) total_length));
			if (iter==(n_iter-1)) {
				message ("rms: %f", sum);
				stat_rms_test = sum;
			}
			if (batch_mode) {
				for (int label=0; label<n_classes; label++) {
					PredictorCluster_BatchAdapt (p[label]);
				}
			}
		}


		
		/*** Testing ***/
		stat_class_err_train = test_model (p, train_data, window_size,
										   delay, cutoff, n_classes, prior );
		message ("Training Classification Error: %f%%%", stat_class_err_train);
		stat_class_err_test = test_model (p, test_data, window_size,
										  delay, cutoff, n_classes, prior );
		message ("Testing Classification Error: %f%%%", stat_class_err_test);

		printf ("%d %d %d %d ", (big_iter+1)*n_iter, n_hidden, n_hidden_b, rbf_layer);
		printf ("%d %d %d ", delay, look_ahead, embedding_dimension);
		printf ("%d ", n_clusters);
		printf ("%f %d %f %f %f %f", learning_rate, prior, stat_rms_train, stat_rms_test, stat_class_err_train, stat_class_err_test);
		printf (" #TOPLOT\n"); //mark this line for grepping
	} /* bigiter */  

	/** Cleaning up **/
	for (int i=0; i<n_predictors; i++) {
		DeletePredictorCluster (p[i]);
	}
	FreeM (p);
	FreeIOExampleList (train_data);
	FreeIOExampleList (test_data);

	message ("\nExiting\n");
	return 0;
}

real test_model (PredictorCluster** p, IOExampleList* test_data, int window_size, int delay, int cutoff, int n_classes, int prior )
{
	int class_err = 0;

	SAMPLE_BUFFER* error_buffer = NULL;
	if (prior) {
		error_buffer = NewSampleBuffer ();
	}

	for (SelectFirstExample (test_data); test_data->current;
		 SelectNextExample (test_data)) {
    
		real* data = test_data->input->x;
		int max_len = test_data->input->length - cutoff;
		int err_len = max_len + delay;
		int label = test_data->target;//((int) (*(test_target[j]->x)));
		assert (( (label>=0) && (label<n_classes) ));

		if (prior) {
			RequestBufferLength (error_buffer, err_len);
			for (int i=0; i<max_len; i+=delay) {
				PredictorCluster_Input (p[n_classes], &data[i]);
				real* y = PredictorCluster_GetOutput (p[n_classes]);
				real* b = error_buffer->x;
				for (int t=0; t<delay; t++) {
					int index = window_size+i-p[n_classes]->T+t+1;
					if ((i+t)>=error_buffer->length) {
						Serror ("Buffer Length %ld, trying to write to %d+%d=%d (c:%d)\n", error_buffer->length, i,t,i+t,cutoff);
					}
					b[i+t] = data[index] - y[t];
				}
			}
			max_len = max_len - cutoff;
			data = error_buffer->x;
		}
		real* sum = AllocM(real, n_classes);
		real minerr;
		int minerr_i = 0;
		for (int cl=0; cl<n_classes; cl++) {
			sum[cl] = 0.0f;
			for (int i=0; i<max_len; i+=delay) {
				//if (states)
				//printf ("%d ", PredictorCluster_ArgMinErr(p[cl], &data[i]));
				sum[cl] += PredictorCluster_Train (p[cl], &data[i]);
			}
			//if (states)
			//printf ("##\n");
			sum[cl] = sqrt (sum[cl] / ((real) max_len));
		}
		minerr = sum[0];
		minerr_i = 0;
		//message ("Target: %d %d", label, max_len);
		for (int cl=0; cl<n_classes; cl++) {
			//message ("%d %f", cl, sum[cl]);
			if (sum[cl] < minerr) {
				minerr = sum[cl];
				minerr_i = cl;
			}
		}
		if (minerr_i!=label) {
			class_err ++;
		}
	}

	if (error_buffer) {
		FreeSampleBuffer (error_buffer);
	}


	return 100.0f*((real)class_err)/((real)test_data->n);
}
