#include <stdio.h>
#include <stdlib.h>
#include "string_utils.h"
#include "debug.h"
#include <assert.h>

int ConvertStringsToBinary(char* dir_name, char*  labels, char* preamble, char* suffix, char** file);
int ConvertSingleFile(char* fn, FILE* fout_stream);

int main (int argc, char** argv) {
  char dir_name[] = "/home/olethros/projects/common/src/data/eeg/";
  char train_labels[] = "train_labels.dat";
  char test_labels[] = "test_labels.dat";
  char train_preamble[] = "erpTrain";
  char test_preamble[] = "erpTest";
  char name_suffix[] = "_mean.dat";
  char* train_file[] = {"train.list", "train_right.list"};
  char* test_file[] = {"test.list", "test_right.list"};

  
  printf("Converting train data\n");
  ConvertStringsToBinary(dir_name, train_labels, train_preamble, name_suffix, train_file);
  printf("Converting test data\n");
  ConvertStringsToBinary(dir_name, test_labels, test_preamble, name_suffix, test_file);
    
  return 0;
}


int ConvertStringsToBinary(char* dir_name, char*  labels, char* preamble, char* suffix, char** file)
{
  int i = 0;
  int n_examples = 0;
  int retval;
  char* flabels = NULL;
  FILE* fstream = NULL;
  char* fout_name[2];
  FILE* fout_stream[2];
  int class_examples[2] = {0,0};
  for (i=0; i<2; i++) {
    fout_name[i] = strConcat(2, dir_name, file[i]);
    if (!WriteFile(fout_name[i], fout_stream[i])) {
      Serror("Fatal error");
      goto quick_exit;
    }
  }
  
  flabels = strConcat(2, dir_name, labels);

  if (!(ReadFile(flabels, fstream))) {
    Serror("Could not open file\n");
    goto quick_exit;
  }

  do {
    int val = 0;
    char* finput_name = NULL;
		char* foutput_name = NULL;
    char cbuf[256];
    retval = fscanf(fstream, "%d", &val);
		if ((retval)&&(retval!=EOF)) {
			n_examples++;
			if (val==-1) val = 0;
			class_examples[val]++;
			sprintf(cbuf, "%d", n_examples);
			finput_name = strConcat(4, dir_name, preamble, cbuf, suffix);
			foutput_name = strConcat(4, dir_name, preamble, cbuf, ".bin");
			FILE* fileout = fopen (foutput_name, "w");
			ConvertSingleFile(finput_name, fileout);
			fprintf (fout_stream[0], "%s label%d.lab\n", foutput_name, val);
			fclose (fileout);
			FreeM(finput_name);
		}
  } while ((retval)&&(retval!=EOF));
  
  if (!retval) {
    Swarning("Scanning value file produced bogus result (line %d)",n_examples);
  }
  printf("Wrote %d examples (%d and %d) \n",n_examples,class_examples[0],class_examples[1]);
 quick_exit:

  if (fstream){
    CloseFile(fstream);
  }

  for (i=0; i<2; i++) {
    FreeM(fout_name[i]);
    if (fout_stream[i]) {
      CloseFile(fout_stream[i]);
    }
  }
  
  FreeM(flabels);

  return 0;

}

int ConvertSingleFile(char* fn, FILE* fout_stream)
{
  FILE* fs = NULL;
  float inval;
  short outval;
  int retval;
  printf("Opening file %s\n",fn);

  if (ReadFile(fn, fs)) {
    do {
      retval = fscanf(fs, "%f", &inval);
      if ((retval)&&(retval!=EOF)) {
				inval = 256.0f*inval;
				outval = (short) inval;
				fwrite(&outval, sizeof(short), 1, fout_stream);
      }
    } while ((retval)&&(retval!=EOF));

    if (retval==0) {
      Swarning("There must have been an error reading %s!",fn);
    }
    
    CloseFile(fs);

    return 0;
  }

  Serror("Could not open file %s\n",fn);

  return -1;
}
