// Copyright (C) 2002 Johnny Mariethoz (Johnny.Mariethoz@idiap.ch)
//                and Samy Bengio (bengio@idiap.ch)
//                and Ronan Collobert (collober@iro.umontreal.ca)
//                
//
// This file is part of Torch. Release II.
// [The Ultimate Machine Learning Library]
//
// Torch is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// Torch is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Torch; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef STRING_UTILS_INC
#define STRING_UTILS_INC

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "debug.h"

/** Some simple functions for string operations.

    @author Samy Bengio (bengio@idiap.ch)
    @author Johnny Mariethoz (Johnny.Mariethoz@idiap.ch)
    @author Ronan Collobert (collober@iro.umontreal.ca)
    @author Christos Dimitrakakis (dimitrak@idiap.ch)
*/
//@{
/** The StringBuffer structure stores buffers of strings */
typedef struct StringBuffer_ {
     char* c; //This is the buffer.
     char* string; //This is the string
     unsigned int length; //This is the buffer length
} StringBuffer;

///Make a new stringbuffer
StringBuffer* NewStringBuffer (int length);

///Given a pointer to a stringbuffer pointer, free it and clear it
void FreeStringBuffer (StringBuffer** s);

/** Returns the name of a file without leading pathname.
    (It's not a new string, but a pointer in the given string)
 */
char *strBaseName(char *filename);

/** Returns a fresh copy of the name of a file without suffix.
    (Trailing chars after c) You have to free the memory!
*/
char *strRemoveSuffix(char *filename, char c='.');

/** Returns the concatenation #n# strings.
    The strings are the parameters given after #n#;
    You have to free the memory!
*/
char *strConcat(int n, ...);

/** Prints a message */
void message(const char* msg, ...);
/** Creates a message string dynamically */
char* make_message (const char *fmt, ...);
/** Reads a string dynamically */
StringBuffer* read_string (FILE* f, StringBuffer* s);
/** copies a string */
char* string_copy (char* c);
//@}



#endif
