/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* $Id: tag2num.c,v 1.4 2003/08/11 10:07:22 olethros Exp $ */
/*
  DESCRIPTION
  
  - This simple program is used in conjuction with the DataFile class
  methods, in order to extract symbols from data tables and tag then
  numerically on a per-column basis. Each column is read once and all
  non-numbers are entered into a tag-list. Then conversion occurs.


  SEE ALSO
  - let2num.c
  - datafile.c

*/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include "tag2num.h"
#include "LinkedList.h"



int main(int argc, char* argv[])
{
  char* InFileName = NULL;
  char* OutFileName = NULL;
  DATAFILE* DataFile;
  int column;

  if (argc!=4) {
	fprintf(stderr,"USAGE: replace_letters <infile> <outfile> <column> \n");
	return -1;
  }
  
  InFileName = argv[1];
  OutFileName = argv[2];
  column = atoi(argv[3]);
  
  printf("Replacing column %d of file %s - writing to %s\n",column,InFileName, OutFileName);
  if ((DataFile = ScanDataFile(InFileName))==NULL) {
	Serror("Could Not create datafile placeholder\n");
  }


  if (!WriteFile(OutFileName, DataFile->OutputFile)) {
	Serror ("ERROR opening file\n");
	return -1;
  }

  TagsToNumbers(DataFile, column);

  CloseFile(DataFile->InputFile);
  CloseFile(DataFile->OutputFile);

  FreeM(DataFile);

  return 0;

}




DATAFILE* ScanDataFile(char* FileName) {
  char* StringBuffer = NULL;
  int StringBufferLength = 2048;
  char* err;
  FILE* FileStream;
  DATAFILE* DataFile;

  if (!ReadFile(FileName, FileStream)) {
	Serror ("ERROR opening file\n");
	return NULL;
  }

  printf("Scanning....\n");
  if (!(StringBuffer  = AllocM(char, StringBufferLength))) {
	Serror ("Memory allocation failed\n");
	return NULL;
  }
	  
  if (!(DataFile = AllocM(DATAFILE, 1))) {
	Serror("Memory Allocation Failed\n");
	return NULL;
  }
  

  err = fgets(StringBuffer, StringBufferLength, FileStream);
 
  SetDimensionsFromString(DataFile, StringBuffer);
  if (DataFile->rows == 0) {
	do {
	  DataFile->rows ++;
	  err = fgets(StringBuffer, StringBufferLength, FileStream);
	} while (err);
	rewind(FileStream);
  }

  printf("Data file has %ld columns and %ld rows\n",DataFile->cols, DataFile->rows);

  FreeM(StringBuffer);

  DataFile->InputFile = FileStream;

  return DataFile;

}




int  SetDimensionsFromString(DATAFILE* DataFile, char* StringBuffer) {
  int i;
  char* s;

  i = 0;
  s = strtok(StringBuffer, ", ");
  while (s) {
	
	switch(i) {
	case 0:
	  sscanf(s,"%ld", &DataFile->rows);
	case 1:
	  sscanf(s,"%ld", &DataFile->cols);
	}
	
	printf("%d: %s\n", i, s);
	i++;
	s = strtok(NULL, ", ");
  }

  if (i!=2) {
	printf("Assuming data in tabular format\n");
	DataFile->cols = i;	
	DataFile->rows = 0;
  }
  
  return 0;

}
  



int TagsToNumbers(DATAFILE* DataFile, long column) {
  LINKEDLIST* list = NULL;
  long i;
  char* StringBuffer = NULL;
  int StringBufferLength = 2048;
  char* s;

  if (!(StringBuffer  = AllocM(char, StringBufferLength))) {
	Serror ("Memory allocation failed\n");
	return -1;
  }


  assert(DataFile);
  assert(DataFile->cols);
  assert(DataFile->rows);
  assert(DataFile->InputFile);
  assert(DataFile->OutputFile);

  printf("Beginning conversion\n");

  fprintf(DataFile->OutputFile, "%ld %ld\n", DataFile->rows, DataFile->cols);
  
  printf("Step 1");
  if (!(list = NewList())) {
	Serror ("Error creating list\n");
	return -1;
  }

  printf("2\n");

  //  list = NewList();

  for (i=0; i<DataFile->rows; i++) {

	s = fgets(StringBuffer, StringBufferLength, DataFile->InputFile);
	if (s) {
	  char* token;
	  int i = 0;
	  token = strtok(StringBuffer, ", ");

	  while(token) {
		i++;
		if (i==column) {
		  fprintf(DataFile->OutputFile, "%ld ", GetStringMatchingListItem(list, token)->ID);
		} else {
		  fprintf(DataFile->OutputFile, "%s ",token);
		}
	  token = strtok(NULL, ", ");
	  }
	  fprintf(DataFile->OutputFile, "\n");
		
	}else{
	  Serror("s should not be NULL\n");
	  return -1;
	}

  }	

  DeleteList(list);

	return 0;

}
 

