#include <stdio.h>
#include <sndfile.h>
#include "SampleBuffer.h"

void show_info (const SF_INFO* info);
void show_format (const int format);

int main (int argc, char** argv) 
{
    const char* wav_input = "/com/databases/numbers95/wav_files/NU-92.streetaddr.wav";
    SNDFILE* tmp = NULL;
    SF_INFO file_info;
    SAMPLE_BUFFER* data = NULL;
    printf ("Hello User\nI will try and load a file for you!\n");
    
    if (argc>1) {
        tmp = sf_open (argv[1], SFM_READ, &file_info);
    } else {
        tmp = sf_open (wav_input, SFM_READ, &file_info);
    }
    if (sf_format_check (&file_info)) {
        printf ("File OK\n");
    } else {
        printf ("File SUX\n");
        goto quick_exit;
    }
    show_info (&file_info);
    data = NewSampleBuffer (file_info.frames, 0);
    printf ("Read %lld samples\n", sf_read_float (tmp, data->x, file_info.frames));
    printf ("Length, in seconds: %f\n", ((float) file_info.frames)/((float) file_info.samplerate));
    printf ("ns: %f\n", 1000000.0 * ((float) file_info.frames)/((float) file_info.samplerate));
 quick_exit:
    if (data) {
        printf ("Freeing buffer\n");
        FreeSampleBuffer (data);
    }
    if (tmp) {
        printf ("Closing file\n");
        sf_close (tmp);
    }

    return 0;
}

void show_info (const SF_INFO* info)
{
    printf ("Statistics\n----------\n");
    printf ("n_frames: %lld\n", info->frames);
    printf ("srate: %d\n", info->samplerate);
    printf ("format: ");
    show_format (info->format);
    printf ("\n");
    printf ("sections: %d\n", info->sections);
    printf ("seekable: %d\n", info->seekable);
    printf ("----------\n");
}

void show_format (const int format)
{
    int format_type = format & SF_FORMAT_TYPEMASK;
    int format_sub = format & SF_FORMAT_SUBMASK;
    int format_end = format & SF_FORMAT_ENDMASK;

    switch (format_type) {
    case SF_FORMAT_WAV:
        printf ("WAV ");
        break;
    case SF_FORMAT_AIFF:
        printf ("AIFF ");
        break;
    case SF_FORMAT_AU:
        printf ("AU ");
        break;
    case SF_FORMAT_RAW:
        printf ("RAW ");
        break;
    case SF_FORMAT_PAF:
        printf ("PAF ");
        break;
    case SF_FORMAT_SVX:
        printf ("SVX ");
        break;
    case SF_FORMAT_NIST:
        printf ("NIST ");
        break;
    case SF_FORMAT_VOC:
        printf ("VOC ");
        break;
    case SF_FORMAT_IRCAM:
        printf ("IRCAM ");
        break;
    case SF_FORMAT_W64:
        printf ("W64 ");
        break;
    case SF_FORMAT_MAT4:
        printf ("MAT4 ");
        break;
    case SF_FORMAT_MAT5:
        printf ("MAT5 ");
        break;
    default:
        printf ("Unknown format -");
    }
    switch (format_sub) {
        
    case SF_FORMAT_PCM_U8:
        printf ("PCM U8");
        break;
    case SF_FORMAT_PCM_S8:
        printf ("PCM S8");
        break;
    case SF_FORMAT_PCM_16:
        printf ("PCM S16");
        break;
    case SF_FORMAT_PCM_24:
        printf ("PCM S24");
        break;
    case SF_FORMAT_PCM_32:
        printf ("PCM S32");
        break;
    case SF_FORMAT_FLOAT:
        printf ("Float");
        break;
    case SF_FORMAT_DOUBLE:
        printf ("Double");
        break;
    case SF_FORMAT_ULAW:
        printf ("ULaw");
        break;
    case SF_FORMAT_ALAW:
        printf ("ALaw");
        break;
    default:
        printf ("Unknown ");
    }

    switch (format_end) {
    case SF_ENDIAN_FILE:
        printf (" (end:File)");
        break;
    case SF_ENDIAN_LITTLE:
        printf (" (end:Little)");
        break;
    case SF_ENDIAN_BIG:
        printf (" (end:Big)");
        break;
    case SF_ENDIAN_CPU:
        printf (" (end:CPU)");
        break;
    default:
        printf (" (end: unknown)");
    }
        

}
